#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from flsim.channels.message import Message
from flsim.common.pytest_helper import assertEqual
from flsim.secure_aggregation.secure_aggregator import FixedPointConfig
from flsim.servers.sync_secagg_servers import SyncSecAggServerConfig
from flsim.tests.utils import (
    SampleNet,
    TwoFC,
    model_parameters_equal_to_value,
    create_model_with_value,
)
from hydra.utils import instantiate


class TestSyncSecAggServer:
    def _create_server(self, model, fixedpoint, channel=None):
        return instantiate(
            SyncSecAggServerConfig(fixedpoint=fixedpoint),
            global_model=model,
            channel=channel,
        )

    def test_sync_secagg_server_init(self) -> None:
        """
        Tests whether secure aggregator object is initiated
        """
        model = SampleNet(TwoFC())
        # test secure aggregation with flat FP config
        fixedpoint = FixedPointConfig(num_bytes=2, scaling_factor=100)
        server = self._create_server(model, fixedpoint=fixedpoint)
        assertEqual(len(server._secure_aggregator.converters), 4)
        assertEqual(  # verify an arbitrary layer of the model
            server._secure_aggregator.converters["fc2.bias"].scaling_factor, 100
        )

    def test_secure_aggregator_receive_update_from_client(self):
        """
        Tests whether secure aggregator operations work correctly
        when a model update is received and server model is updated

        Same test as https://fburl.com/code/vp7o2clh
        """
        scaling_factor = 100
        fixedpoint = FixedPointConfig(num_bytes=2, scaling_factor=scaling_factor)
        server = self._create_server(
            SampleNet(create_model_with_value(0)), fixedpoint=fixedpoint
        )
        server.init_round()

        m1_param = 7.2345
        m1_w = 3.0
        model1 = create_model_with_value(m1_param)
        server.receive_update_from_client(Message(SampleNet(model1), weight=m1_w))

        m2_param = -3.45612
        m2_w = 7.0
        model2 = create_model_with_value(m2_param)
        server.receive_update_from_client(Message(SampleNet(model2), weight=m2_w))

        expected_param = float(
            round(m1_param * scaling_factor * m1_w + m2_param * scaling_factor * m2_w)
        )

        server.step()
        mismatched = model_parameters_equal_to_value(
            server.global_model.fl_get_module(),
            -(expected_param / scaling_factor) / (m1_w + m2_w),
        )
        assertEqual(mismatched, "", mismatched)

    def test_secure_aggregator_step_large_range(self):
        """
        Tests whether secure aggeragtion operations work correctly
        when the step() method is called, and when the num_bytes is
        big, so we do not have a possible fixedpoint overflow

        Same test as https://fburl.com/code/qp7qve2m
        """
        scaling_factor = 10
        num_bytes = 4
        global_param = 8.0
        client_param = 2.123
        num_clients = 10

        fixedpoint = FixedPointConfig(
            num_bytes=num_bytes, scaling_factor=scaling_factor
        )
        server = self._create_server(
            SampleNet(create_model_with_value(global_param)), fixedpoint=fixedpoint
        )

        clients = [create_model_with_value(client_param) for _ in range(num_clients)]

        server.init_round()
        for client in clients:
            server.receive_update_from_client(Message(SampleNet(client), weight=1.0))

        expected_param = float(round(global_param - client_param, ndigits=1))

        server.step()
        mismatched = model_parameters_equal_to_value(
            server.global_model.fl_get_module(), expected_param
        )
        assertEqual(mismatched, "", mismatched)

    def test_secure_aggregator_step_small_range(self):
        """
        Tests whether secure aggeragtion operations work correctly
        when the step() method is called, and when the num_bytes is
        small so we have possible fixedpoint overflows

        Same test as https://fburl.com/code/qp7qve2m
        """
        scaling_factor = 100
        num_bytes = 1
        global_param = 8
        client_param = 2.123
        num_clients = 10

        fixedpoint = FixedPointConfig(
            num_bytes=num_bytes, scaling_factor=scaling_factor
        )
        server = self._create_server(
            SampleNet(create_model_with_value(global_param)), fixedpoint=fixedpoint
        )

        clients = [create_model_with_value(client_param) for _ in range(num_clients)]

        server.init_round()
        for client in clients:
            server.receive_update_from_client(Message(SampleNet(client), weight=1.0))

        # update is 5.877. In current fixedpoint setup, it is 127, as 587.7 > 127
        # when we reduce, we convert from fixedpoint to floating point, so 127 -> 1.27
        expected_param = float(global_param - (1.27 * num_clients) / num_clients)

        server.step()
        mismatched = model_parameters_equal_to_value(
            server.global_model.fl_get_module(), expected_param
        )
        assertEqual(mismatched, "", mismatched)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import abc
import copy
from typing import Dict, Iterable

import torch


# TODO: (jesikmin) T55061072 Add test under papaya directory for
# FLModelWithPrivateModules
class FLModelWithPrivateModules(abc.ABC):
    """
    This class emulates the user-private modules in FL by making them reference to
    class-level global attributes.
    The user of this class must make sure the invariance that the user-private
    modules must reference to class-level attributes.

    In federated learning, this emulation should be handled at the following points:
    when a new instance is created, when models are copied to each user, when doing
    forward propagation, when backprop, and when user models are averaged back to the
    server.
    """

    USER_PRIVATE_MODULE_PREFIX = "USER_PRIVATE_MODULE"

    user_private_module_dict: Dict[str, torch.nn.Module] = {}

    @classmethod
    def clear_user_private_module_dict(cls):
        cls.user_private_module_dict.clear()

    @classmethod
    def get_user_private_parameters(cls) -> Iterable[torch.Tensor]:
        """Return emulated mapping that maps each user to her private params."""
        for module in cls.user_private_module_dict.values():
            for param in module.parameters():
                yield param

    def _get_user_private_module_attr_name(self, module_name):
        return f"{self.USER_PRIVATE_MODULE_PREFIX}_{module_name}"

    def _maybe_set_up_user_private_modules(self, forced: bool = False):
        """
        Set an instance's private modules to class attributes to share among
        all users. This function runs only when all user-private attributes
        have been set.
        """
        if not forced:
            for module_name in self._get_user_private_module_names():
                # The user-private modules may not be set during component creation.
                if not hasattr(self, module_name) or getattr(self, module_name) is None:
                    return

        # Initialize the class attributes if not exist.
        for module_name in self._get_user_private_module_names():
            if module_name not in self.user_private_module_dict:
                self.user_private_module_dict[module_name] = getattr(self, module_name)

            # Replace instance-based private attributes with the class attributes.
            # for module_name in self._get_user_private_module_names():
            # Remove instance version if not removed.
            if hasattr(self, module_name):
                delattr(self, module_name)

            setattr(
                self,
                self._get_user_private_module_attr_name(module_name),
                self.user_private_module_dict[module_name],
            )

    def _set_forward_hooks(self):
        """Set forward hooks to reuse the forward() of the parent class.
        The pre-forward hook changes the name of the user-private parameters
        back to the original ones to reuse the forward() function of the parent
        class. The forward hook changes the name back to have the
        USER_PRIVATE_MODULE_PREFIX.
        """

        def set_user_private_modules(module, inputs):
            for key in module._get_user_private_module_names():
                setattr(module, key, module.user_private_module_dict[key])

        def remove_user_private_modules(module, inputs, outputs):
            for key in module._get_user_private_module_names():
                delattr(module, key)

        self.register_forward_pre_hook(set_user_private_modules)
        self.register_forward_hook(remove_user_private_modules)

    def __deepcopy__(self, memo):
        orig_deepcopy_method = self.__deepcopy__
        self.__deepcopy__ = None

        # Don't want to copy the user-private modules which point to the
        # class-level attributes.
        for module_name in self._get_user_private_module_names():
            delattr(self, self._get_user_private_module_attr_name(module_name))

        cp = copy.deepcopy(self, memo)

        # Re-set-up the user-private params to the class-level attributes.
        self._maybe_set_up_user_private_modules(forced=True)
        cp._maybe_set_up_user_private_modules(forced=True)

        self.__deepcopy__ = orig_deepcopy_method
        return cp

    def get_user_private_attr(self, module_name):
        return getattr(self, self._get_user_private_module_attr_name(module_name))

    @classmethod
    @abc.abstractmethod
    def _get_user_private_module_names(cls) -> Iterable[str]:
        """Return an iterable of the modules of the class to be private."""
        pass

    def federated_state_dict(self):
        """Return a state dict of federated modules."""
        state_dict = self.state_dict()
        # Do not copy user private param modules.
        for key in state_dict.keys():
            if key.startswith(self.USER_PRIVATE_MODULE_PREFIX):
                del state_dict[key]
        return state_dict

    def load_federated_state_dict(self, state_dict: Dict):
        """Load from a state dict of federated modules."""
        # pyre-fixme[16]: `FLModelWithPrivateModules` has no attribute
        #  `load_state_dict`.
        missing_keys, unexpected_keys = self.load_state_dict(
            state_dict=state_dict, strict=False
        )
        assert len(unexpected_keys) == 0, "There should be no unexpected keys"
        for key in missing_keys:
            assert key.startswith(
                self.USER_PRIVATE_MODULE_PREFIX
            ), f"Missing non-user-private parameter {key}"
        return missing_keys, unexpected_keys



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import abc
from typing import Any

import torch.nn as nn
from flsim.interfaces.batch_metrics import IFLBatchMetrics


class IFLModel(abc.ABC):
    @abc.abstractmethod
    def fl_forward(self, batch: Any) -> IFLBatchMetrics:
        pass

    @abc.abstractmethod
    def fl_create_training_batch(self, **kwargs) -> Any:
        pass

    @abc.abstractmethod
    def fl_get_module(self) -> nn.Module:
        pass

    @abc.abstractmethod
    def fl_cuda(self) -> None:
        pass

    @abc.abstractmethod
    def get_eval_metrics(self, batch: Any) -> IFLBatchMetrics:
        pass

    @abc.abstractmethod
    def get_num_examples(self, batch: Any) -> int:
        pass



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import math
from collections import Counter

import torch
from flsim.active_user_selectors.diverse_user_selector import (
    DiverseUserSelectorUtils,
    DiversityMaximizingUserSelector,
    DiversityMaximizingUserSelectorConfig,
    DiversityStatisticsReportingUserSelector,
    DiversityStatisticsReportingUserSelectorConfig,
)
from flsim.common.diversity_metrics import DiversityMetricType
from flsim.common.pytest_helper import assertEqual, assertTrue, assertIsInstance
from flsim.utils.sample_model import DummyAlphabetFLModel, LinearFLModel
from flsim.utils.tests.helpers.test_data_utils import (
    DummyAlphabetDataset,
    NonOverlappingDataset,
    RandomDataset,
)
from hydra.utils import instantiate


class TestDiverseUserSelector:

    # Since diversity statistics reporting selector chooses uniformly randomly, replicate the tests
    # for that class. select_diverse_cohort is tested separately.
    def test_diversity_statistics_reporting_user_selector(self):
        null_selector = instantiate(
            DiversityStatisticsReportingUserSelectorConfig(num_candidate_cohorts=2)
        )
        assertIsInstance(null_selector, DiversityStatisticsReportingUserSelector)
        # This user selector class requires a global model and data provider.
        users = tuple(range(5))

        shard_size, local_batch_size = 4, 4
        dummy_dataset = DummyAlphabetDataset()
        dummy_model = DummyAlphabetFLModel()
        data_provider, _ = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, shard_size, local_batch_size, dummy_model
        )

        uniform_selector_1 = instantiate(
            DiversityStatisticsReportingUserSelectorConfig(
                user_selector_seed=12345, num_candidate_cohorts=2
            )
        )
        uniform_selector_2 = instantiate(
            DiversityStatisticsReportingUserSelectorConfig(
                user_selector_seed=12345, num_candidate_cohorts=2
            )
        )
        # Repeat, since it can't be done with many users as test_uniformly_random_user_selector
        num_trials = 25
        for _ in range(num_trials):
            selection_1 = uniform_selector_1.get_user_indices(
                num_total_users=len(users),
                users_per_round=1,
                data_provider=data_provider,
                global_model=dummy_model,
                epoch=0,
            )
            selection_2 = uniform_selector_2.get_user_indices(
                num_total_users=len(users),
                users_per_round=1,
                data_provider=data_provider,
                global_model=dummy_model,
                epoch=0,
            )
            assertEqual(selection_1, selection_2)

    def test_diversity_maximizing_user_selector(self):
        # Check that the selector actually chooses the cohort with maximum GD.
        # For cohort size 3 our of 4 users, there are 4 possible cohorts.
        # With 100 candidate cohorts we will try all 4 with very high probability.
        num_total_users, users_per_round = 4, 3
        num_data_per_user, dim_data = 6, 40
        shard_size, local_batch_size = 6, 6
        num_candidates = 100
        available_users = range(num_total_users)

        null_selector = instantiate(
            DiversityMaximizingUserSelectorConfig(num_candidate_cohorts=num_candidates)
        )
        assertIsInstance(null_selector, DiversityMaximizingUserSelector)

        linear_model = LinearFLModel(D_in=dim_data, D_out=1)
        random_dataset = RandomDataset(
            num_users=num_total_users,
            num_data_per_user=num_data_per_user,
            dim_data=dim_data,
        )
        data_provider, _ = RandomDataset.create_data_provider_and_loader(
            random_dataset, shard_size, local_batch_size, linear_model
        )

        diverse_cohort = null_selector.get_user_indices(
            num_total_users=num_total_users,
            users_per_round=users_per_round,
            data_provider=data_provider,
            global_model=linear_model,
            epoch=0,
        )

        max_diversity = DiverseUserSelectorUtils.calculate_diversity_metrics(
            data_provider=data_provider,
            global_model=linear_model,
            user_indices=diverse_cohort,
        )

        available_users_set = set(available_users)
        # Enumerate through each possible cohort, show it doesn't have greater Diversity
        for i in available_users:
            cohort = list(available_users_set.difference({i}))

            cohort_diversity = DiverseUserSelectorUtils.calculate_diversity_metrics(
                data_provider=data_provider,
                global_model=linear_model,
                user_indices=cohort,
            )

            assertTrue(cohort_diversity <= max_diversity)

        # Test epochs_before_active as is done in test_high_loss_user_selector()
        selector = instantiate(
            DiversityMaximizingUserSelectorConfig(epochs_before_active=1)
        )

        shard_size, local_batch_size = 10, 4
        dummy_dataset = DummyAlphabetDataset(num_rows=11)
        dummy_model = DummyAlphabetFLModel()
        data_provider, _ = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, shard_size, local_batch_size, dummy_model
        )

        selections = [
            selector.get_user_indices(
                num_total_users=2,
                users_per_round=1,
                data_provider=data_provider,
                global_model=dummy_model,
                epoch=0,
            )[0]
            for i in range(1000)
        ]
        counts = Counter(selections)
        assertTrue(counts[0] > 400)
        assertTrue(counts[0] < 600)

    def test_user_selector_diversity_metric_type(self):
        # Test that the created user selector uses the specified diversity metric

        metric_type_strings = [
            "gradient_diversity",
            "orthogonality",
            "delta_norm_sq",
            "sum_client_delta_norm_sq",
            "sum_client_delta_mutual_angles",
        ]
        metric_types = [
            DiversityMetricType.gradient_diversity,
            DiversityMetricType.orthogonality,
            DiversityMetricType.delta_norm_sq,
            DiversityMetricType.sum_client_delta_norm_sq,
            DiversityMetricType.sum_client_delta_mutual_angles,
        ]

        for metric_str, metric_type in zip(metric_type_strings, metric_types):
            user_selector = instantiate(
                DiversityMaximizingUserSelectorConfig(diversity_metric_type=metric_str)
            )
            assertEqual(user_selector.cfg.diversity_metric_type, metric_type)


class TestDiverseUserSelectorUtils:

    tolerence = 1e-5

    def test_calculate_diversity_metrics(self):
        # We construct nonoverlapping data for each user so that the gradient diversity will be 1
        num_total_users, users_per_round = 10, 3
        num_nonzeros_per_user, num_data_per_user = 4, 6
        shard_size, local_batch_size = 6, 6
        linear_model = LinearFLModel(
            D_in=num_total_users * num_nonzeros_per_user, D_out=1
        )
        nonoverlap_dataset = NonOverlappingDataset(
            num_users=num_total_users,
            num_nonzeros_per_user=num_nonzeros_per_user,
            num_data_per_user=num_data_per_user,
        )
        data_provider, _ = NonOverlappingDataset.create_data_provider_and_loader(
            nonoverlap_dataset, shard_size, local_batch_size, linear_model
        )

        # Check GD for all users then for random cohorts; ensure GD=1 for each
        available_users = range(num_total_users)
        rng = torch.Generator()
        diversity_metrics = DiverseUserSelectorUtils.calculate_diversity_metrics(
            data_provider, linear_model, available_users
        )
        assertTrue(
            math.isclose(diversity_metrics.gradient_diversity, 1.0, rel_tol=1e-04)
        )

        while len(available_users) > 0:
            selected_indices = torch.multinomial(
                torch.ones(len(available_users), dtype=torch.float),
                min(users_per_round, len(available_users)),
                replacement=False,
                generator=rng,
            ).tolist()
            user_indices = [available_users[idx] for idx in selected_indices]
            available_users = [
                idx for idx in available_users if idx not in user_indices
            ]
            diversity_metrics = DiverseUserSelectorUtils.calculate_diversity_metrics(
                data_provider, linear_model, user_indices
            )
            assertTrue(
                math.isclose(diversity_metrics.gradient_diversity, 1.0, rel_tol=1e-04)
            )

        # By the definition of GD, it will always be at least 1/num_users.
        num_total_users, users_per_round = 10, 3
        num_data_per_user, dim_data = 6, 40
        shard_size, local_batch_size = 6, 6
        linear_model = LinearFLModel(
            D_in=num_total_users * num_nonzeros_per_user, D_out=1
        )
        random_dataset = RandomDataset(
            num_users=num_total_users,
            num_data_per_user=num_data_per_user,
            dim_data=dim_data,
        )
        data_provider, _ = RandomDataset.create_data_provider_and_loader(
            random_dataset, shard_size, local_batch_size, linear_model
        )

        # Check GD for all users then for random cohorts; ensure GD>=1/num_users for each
        available_users = range(num_total_users)
        rng = torch.Generator()
        diversity_metrics = DiverseUserSelectorUtils.calculate_diversity_metrics(
            data_provider, linear_model, available_users
        )
        assertTrue(diversity_metrics.gradient_diversity >= 1.0 / num_total_users)

        while len(available_users) > 0:
            selected_indices = torch.multinomial(
                torch.ones(len(available_users), dtype=torch.float),
                min(users_per_round, len(available_users)),
                replacement=False,
                generator=rng,
            ).tolist()
            user_indices = [available_users[idx] for idx in selected_indices]

            diversity_metrics = DiverseUserSelectorUtils.calculate_diversity_metrics(
                data_provider, linear_model, user_indices
            )
            assertTrue(
                diversity_metrics.gradient_diversity
                >= 1.0 / min(num_total_users, len(available_users))
            )

            available_users = [
                idx for idx in available_users if idx not in user_indices
            ]

    def test_select_diverse_cohort(self):

        # Use dataset in which the gradient diversity is 1 for any cohort and
        # Check max, avg, and min GD = 1
        num_total_users, users_per_round = 10, 3
        num_nonzeros_per_user, num_data_per_user = 4, 6
        shard_size, local_batch_size = 6, 6
        num_candidate_cohorts = 5
        loss_reduction_type = "mean"
        client_gradient_scaling = "sum"
        diversity_metric_type = DiversityMetricType.gradient_diversity
        maximize_metric = True
        rel_tol = 1e-04
        available_users = range(num_total_users)
        rng = torch.Generator()

        linear_model = LinearFLModel(
            D_in=num_total_users * num_nonzeros_per_user, D_out=1
        )
        nonoverlap_dataset = NonOverlappingDataset(
            num_users=num_total_users,
            num_nonzeros_per_user=num_nonzeros_per_user,
            num_data_per_user=num_data_per_user,
        )
        data_provider, _ = NonOverlappingDataset.create_data_provider_and_loader(
            nonoverlap_dataset, shard_size, local_batch_size, linear_model
        )

        (_, diversity_statistics,) = DiverseUserSelectorUtils.select_diverse_cohort(
            data_provider=data_provider,
            global_model=linear_model,
            users_per_round=users_per_round,
            available_users=available_users,
            rng=rng,
            num_search_samples=num_candidate_cohorts,
            maximize_metric=maximize_metric,
            loss_reduction_type=loss_reduction_type,
            client_gradient_scaling=client_gradient_scaling,
            diversity_metric_type=diversity_metric_type,
        )

        assertTrue(
            math.isclose(diversity_statistics.maximum_metric, 1.0, rel_tol=rel_tol)
        )
        assertTrue(
            math.isclose(diversity_statistics.average_metric, 1.0, rel_tol=rel_tol)
        )
        assertTrue(
            math.isclose(diversity_statistics.minimum_metric, 1.0, rel_tol=rel_tol)
        )

        # On random data, check diversity statistics for all users then for random cohorts. Ensure that
        # there is separation between the max, average, and min when not all users are in the cohort.
        num_total_users, users_per_round = 10, 3
        num_data_per_user, dim_data = 6, 40
        shard_size, local_batch_size = 6, 6
        num_candidate_cohorts = 5
        loss_reduction_type = "mean"
        client_gradient_scaling = "sum"
        diversity_metric_type = DiversityMetricType.gradient_diversity
        maximize_metric = True
        available_users = range(num_total_users)

        linear_model = LinearFLModel(D_in=dim_data, D_out=1)
        random_dataset = RandomDataset(
            num_users=num_total_users,
            num_data_per_user=num_data_per_user,
            dim_data=dim_data,
        )
        data_provider, _ = RandomDataset.create_data_provider_and_loader(
            random_dataset, shard_size, local_batch_size, linear_model
        )
        rng = torch.Generator()

        (
            diverse_cohort,
            diversity_statistics,
        ) = DiverseUserSelectorUtils.select_diverse_cohort(
            data_provider=data_provider,
            global_model=linear_model,
            users_per_round=num_total_users,
            available_users=available_users,
            rng=rng,
            num_search_samples=num_candidate_cohorts,
            maximize_metric=maximize_metric,
            loss_reduction_type=loss_reduction_type,
            client_gradient_scaling=client_gradient_scaling,
            diversity_metric_type=diversity_metric_type,
        )
        assertEqual(set(diverse_cohort), set(available_users))
        assertTrue(
            math.isclose(
                diversity_statistics.maximum_metric,
                diversity_statistics.average_metric,
                rel_tol=rel_tol,
            )
        )
        assertTrue(
            math.isclose(
                diversity_statistics.average_metric,
                diversity_statistics.minimum_metric,
                rel_tol=rel_tol,
            )
        )

        while len(available_users) > 0:

            (
                diverse_cohort,
                diversity_statistics,
            ) = DiverseUserSelectorUtils.select_diverse_cohort(
                data_provider=data_provider,
                global_model=linear_model,
                users_per_round=users_per_round,
                available_users=available_users,
                rng=rng,
                num_search_samples=num_candidate_cohorts,
                maximize_metric=maximize_metric,
                loss_reduction_type=loss_reduction_type,
                client_gradient_scaling=client_gradient_scaling,
                diversity_metric_type=diversity_metric_type,
            )

            # Check that the chosen cohort is valid
            assertTrue(set(diverse_cohort) <= set(available_users))

            assertTrue(
                diversity_statistics.minimum_metric
                >= 1.0 / min(num_total_users, len(available_users))
            )

            # The gradient diversity should be equal to the maximum in the statistics
            diverse_cohort_diversity = (
                DiverseUserSelectorUtils.calculate_diversity_metrics(
                    data_provider=data_provider,
                    global_model=linear_model,
                    user_indices=diverse_cohort,
                    loss_reduction_type=loss_reduction_type,
                    client_gradient_scaling=client_gradient_scaling,
                    diversity_metric_type=diversity_metric_type,
                )
            )
            assertTrue(
                math.isclose(
                    diverse_cohort_diversity.gradient_diversity,
                    diversity_statistics.maximum_metric,
                    rel_tol=rel_tol,
                )
            )

            if len(available_users) > users_per_round:
                assertEqual(len(diverse_cohort), users_per_round)
                assertTrue(
                    diversity_statistics.maximum_metric
                    > diversity_statistics.average_metric
                )
                assertTrue(
                    diversity_statistics.average_metric
                    > diversity_statistics.minimum_metric
                )
            else:
                assertEqual(len(diverse_cohort), len(available_users))
                assertTrue(
                    math.isclose(
                        diversity_statistics.maximum_metric,
                        diversity_statistics.average_metric,
                        rel_tol=rel_tol,
                    )
                )
                assertTrue(
                    math.isclose(
                        diversity_statistics.average_metric,
                        diversity_statistics.minimum_metric,
                        rel_tol=rel_tol,
                    )
                )
            available_users = [
                idx for idx in available_users if idx not in diverse_cohort
            ]

        # Try this again with minimizing the diversity metric
        num_total_users, users_per_round = 10, 3
        num_data_per_user, dim_data = 6, 40
        shard_size, local_batch_size = 6, 6
        num_candidate_cohorts = 5
        loss_reduction_type = "mean"
        client_gradient_scaling = "sum"
        diversity_metric_type = DiversityMetricType.gradient_diversity
        maximize_metric = False
        available_users = range(num_total_users)

        linear_model = LinearFLModel(D_in=dim_data, D_out=1)
        random_dataset = RandomDataset(
            num_users=num_total_users,
            num_data_per_user=num_data_per_user,
            dim_data=dim_data,
        )
        data_provider, _ = RandomDataset.create_data_provider_and_loader(
            random_dataset, shard_size, local_batch_size, linear_model
        )
        rng = torch.Generator()

        (
            diverse_cohort,
            diversity_statistics,
        ) = DiverseUserSelectorUtils.select_diverse_cohort(
            data_provider=data_provider,
            global_model=linear_model,
            users_per_round=num_total_users,
            available_users=available_users,
            rng=rng,
            num_search_samples=num_candidate_cohorts,
            maximize_metric=maximize_metric,
            loss_reduction_type=loss_reduction_type,
            client_gradient_scaling=client_gradient_scaling,
            diversity_metric_type=diversity_metric_type,
        )
        assertEqual(set(diverse_cohort), set(available_users))
        assertTrue(
            math.isclose(
                diversity_statistics.maximum_metric,
                diversity_statistics.average_metric,
                rel_tol=rel_tol,
            )
        )
        assertTrue(
            math.isclose(
                diversity_statistics.average_metric,
                diversity_statistics.minimum_metric,
                rel_tol=rel_tol,
            )
        )

        while len(available_users) > 0:

            (
                diverse_cohort,
                diversity_statistics,
            ) = DiverseUserSelectorUtils.select_diverse_cohort(
                data_provider=data_provider,
                global_model=linear_model,
                users_per_round=users_per_round,
                available_users=available_users,
                rng=rng,
                num_search_samples=num_candidate_cohorts,
                maximize_metric=maximize_metric,
                loss_reduction_type=loss_reduction_type,
                client_gradient_scaling=client_gradient_scaling,
                diversity_metric_type=diversity_metric_type,
            )

            # Check that the chosen cohort is valid
            assertTrue(set(diverse_cohort) <= set(available_users))

            assertTrue(
                diversity_statistics.minimum_metric
                >= 1.0 / min(num_total_users, len(available_users))
            )

            # The gradient diversity should be equal to the minimium in the statistics
            diverse_cohort_diversity = (
                DiverseUserSelectorUtils.calculate_diversity_metrics(
                    data_provider=data_provider,
                    global_model=linear_model,
                    user_indices=diverse_cohort,
                    loss_reduction_type=loss_reduction_type,
                    client_gradient_scaling=client_gradient_scaling,
                    diversity_metric_type=diversity_metric_type,
                )
            )
            assertTrue(
                math.isclose(
                    diverse_cohort_diversity.gradient_diversity,
                    diversity_statistics.minimum_metric,
                    rel_tol=rel_tol,
                )
            )

            if len(available_users) > users_per_round:
                assertEqual(len(diverse_cohort), users_per_round)
                assertTrue(
                    diversity_statistics.maximum_metric
                    > diversity_statistics.average_metric
                )
                assertTrue(
                    diversity_statistics.average_metric
                    > diversity_statistics.minimum_metric
                )
            else:
                assertEqual(len(diverse_cohort), len(available_users))
                assertTrue(
                    math.isclose(
                        diversity_statistics.maximum_metric,
                        diversity_statistics.average_metric,
                        rel_tol=rel_tol,
                    )
                )
                assertTrue(
                    math.isclose(
                        diversity_statistics.average_metric,
                        diversity_statistics.minimum_metric,
                        rel_tol=rel_tol,
                    )
                )
            available_users = [
                idx for idx in available_users if idx not in diverse_cohort
            ]



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import collections
import copy
import math

import torch
import torch.nn as nn
from flsim.utils.fl.common import FLModelParamUtils
from flsim.utils.fl.personalized_model import FLModelWithPrivateModules
from flsim.utils.tests.helpers.test_models import (
    FCModel,
    LinearRegression,
    PersonalizedLinearRegression,
)
from flsim.utils.tests.helpers.test_utils import FLTestUtils
from libfb.py import testutil


PRIVATE_SLOPE_MODULE_NAME = FLModelWithPrivateModules.USER_PRIVATE_MODULE_PREFIX + "_a"


class FLModelParamUtilsTest(testutil.BaseFacebookTestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_get_state_dict(self) -> None:
        model = LinearRegression()
        self.assertEqual(
            set(FLModelParamUtils.get_state_dict(model, False).keys()), {"a", "b"}
        )
        self.assertEqual(
            set(FLModelParamUtils.get_state_dict(model, True).keys()), {"a", "b"}
        )

        personalized_model = PersonalizedLinearRegression()
        self.assertEqual(
            set(FLModelParamUtils.get_state_dict(personalized_model, False).keys()),
            {PRIVATE_SLOPE_MODULE_NAME, "b"},
        )
        self.assertEqual(
            set(FLModelParamUtils.get_state_dict(personalized_model, True).keys()),
            {"b"},
        )

    def test_load_state_dict(self) -> None:
        personalized_model = PersonalizedLinearRegression()

        state_dict = collections.OrderedDict()
        state_dict[PRIVATE_SLOPE_MODULE_NAME] = torch.tensor([1.0])
        state_dict["b"] = torch.tensor([0.5])

        FLModelParamUtils.load_state_dict(personalized_model, state_dict, False)
        self.assertEqual(
            dict(FLModelParamUtils.get_state_dict(personalized_model, False)),
            dict(state_dict),
        )
        # load_state_dict should work if non-private modules were given with
        # only_federated_params set as True
        state_dict_without_private_module = collections.OrderedDict()
        state_dict_without_private_module["b"] = torch.tensor([0.3])
        FLModelParamUtils.load_state_dict(
            personalized_model, state_dict_without_private_module, True
        )
        self.assertEqual(
            dict(FLModelParamUtils.get_state_dict(personalized_model, False)),
            {PRIVATE_SLOPE_MODULE_NAME: torch.tensor([1.0]), "b": torch.tensor([0.3])},
        )
        # throws when unexpected key is provided
        state_dict["c"] = torch.tensor([0.0])
        with self.assertRaises(AssertionError):
            FLModelParamUtils.load_state_dict(personalized_model, state_dict, True)
        # throws when non-private (i.e. federated module) is missing
        state_dict_with_missing_non_private_module = collections.OrderedDict()
        state_dict_with_missing_non_private_module["a"] = torch.tensor([1.0])
        with self.assertRaises(AssertionError):
            FLModelParamUtils.load_state_dict(
                personalized_model, state_dict_with_missing_non_private_module, True
            )

    def test_zero_weights(self) -> None:
        personalized_model = PersonalizedLinearRegression()
        FLModelParamUtils.load_state_dict(
            personalized_model,
            collections.OrderedDict(
                [
                    (PRIVATE_SLOPE_MODULE_NAME, torch.tensor([2.0])),
                    ("b", torch.tensor([1.0])),
                ]
            ),
            False,
        )
        FLModelParamUtils.zero_weights(personalized_model, True)
        self.assertEqual(
            dict(FLModelParamUtils.get_state_dict(personalized_model, False)),
            {PRIVATE_SLOPE_MODULE_NAME: torch.tensor([2.0]), "b": torch.tensor([0.0])},
        )
        FLModelParamUtils.zero_weights(personalized_model)
        self.assertEqual(
            dict(FLModelParamUtils.get_state_dict(personalized_model, False)),
            {PRIVATE_SLOPE_MODULE_NAME: torch.tensor([0.0]), "b": torch.tensor([0.0])},
        )

    def test_get_trainable_params(self):
        fc_model = FCModel()
        self.assertEqual(len(list(FLModelParamUtils.get_trainable_params(fc_model))), 6)

    def test_get_num_trainable_params(self):
        fc_model = FCModel()
        self.assertEqual(
            FLModelParamUtils.get_num_trainable_params(fc_model),
            10 * 5 + 5 * 3 + 3 * 1 + 5 + 3 + 1,
        )

    def test_get_gradient_l2_norm_raw(self):
        fc_model = FCModel()
        # set all gradients to 0, l2 norm should be zero
        for p in FLModelParamUtils.get_trainable_params(fc_model):
            p.grad = torch.zeros_like(p)
        self.assertEqual(FLModelParamUtils.get_gradient_l2_norm_raw(fc_model), 0.0)

        # set all gradients to 1, non-normalized l2 norm should be = sqrt(#params)
        num_trainable_params = FLModelParamUtils.get_num_trainable_params(fc_model)
        for p in FLModelParamUtils.get_trainable_params(fc_model):
            p.grad = torch.ones_like(p)
        self.assertAlmostEqual(
            FLModelParamUtils.get_gradient_l2_norm_raw(fc_model),
            math.sqrt(num_trainable_params),
            delta=1e-4,
        )

        # all gradients are std-normal-random, normalized grad norm = 1
        torch.manual_seed(1)
        for p in FLModelParamUtils.get_trainable_params(fc_model):
            p.grad = torch.randn_like(p)
        self.assertAlmostEqual(
            FLModelParamUtils.get_gradient_l2_norm_normalized(fc_model), 1, delta=1e-1
        )

    def test_model_linear_comb(self):
        """Test that computing linear comibination works for a model"""
        FLTestUtils.compare_model_linear_comb(FCModel(), FCModel())

    def test_gradient_reconstruction(self):
        """Test that gradient reconstruction works with a model.
        Create model, run some operations on it.
        """
        model, copy_model, reconstructed_grad = FCModel(), FCModel(), FCModel()
        FLTestUtils.compare_gradient_reconstruction(
            model, copy_model, reconstructed_grad
        )

    def test_fed_async_aggregation_with_weights(self):
        """Test that weights work for FedAsync aggregation"""
        torch.manual_seed(1)
        num_models = 4
        models = [FCModel() for i in range(num_models)]
        temp_model = copy.deepcopy(models[0])
        # verify that 0 weights work as expected
        FLModelParamUtils.average_models(models, temp_model, [0, 0, 0, 1])
        self.assertTrue(
            FLModelParamUtils.get_mismatched_param([temp_model, models[3]]) == ""
        )
        # verify that equal weights work as expected
        FLModelParamUtils.average_models(models, temp_model, [1, 1, 1, 1])
        temp_model_no_wts = copy.deepcopy(models[0])
        FLModelParamUtils.average_models(models, temp_model_no_wts)
        self.assertTrue(
            FLModelParamUtils.get_mismatched_param([temp_model, temp_model_no_wts])
            == ""
        )
        # verify that unequal weights work as expected
        temp_model_1 = copy.deepcopy(models[0])
        FLModelParamUtils.average_models(models, temp_model_1, [1, 1, 2, 2])
        temp_model_2 = copy.deepcopy(models[0])
        FLModelParamUtils.average_models(models, temp_model_2, [2, 2, 1, 1])
        temp_model_3 = copy.deepcopy(models[0])
        FLModelParamUtils.average_models([temp_model_1, temp_model_2], temp_model_3)
        temp_model_4 = copy.deepcopy(models[0])
        FLModelParamUtils.average_models(models, temp_model_4, [1, 1, 1, 1])

        mismatched_param = FLModelParamUtils.get_mismatched_param(
            [temp_model_3, temp_model_4], 1e-6
        )
        self.assertTrue(
            mismatched_param == "",
            (
                f"Mismatched param name: {mismatched_param}\n"
                f"temp_model_3:{temp_model_3}\n"
                f"temp_model_4:{temp_model_4}\n",
                f"total_difference:{self._compute_difference_in_norm(temp_model_3, temp_model_4)}",
            ),
        )

    def _compute_difference_in_norm(
        self, model1: torch.nn.Module, model2: torch.nn.Module
    ):
        total_difference = 0.0
        for (parameter1, parameter2) in zip(model1.parameters(), model2.parameters()):
            total_difference += torch.norm(parameter1.data - parameter2.data)
        return total_difference

    def test_simple_model_copy(self):
        """Test that FedAsync aggregation works for a simple Model"""
        num_models = 4
        orig_models = [FCModel() for i in range(num_models)]
        FLTestUtils.average_and_verify_models(orig_models)

    def test_debug_model_norm(self):
        fc_model = FCModel()
        for p in fc_model.parameters():
            torch.nn.init.constant_(p, 0.0)
        self.assertEqual(FLModelParamUtils.debug_model_norm(fc_model), 0)
        for p in fc_model.parameters():
            p.data.fill_(1.0)
        self.assertEqual(
            FLModelParamUtils.debug_model_norm(fc_model),
            FLModelParamUtils.get_num_trainable_params(fc_model),
        )

    def test_set_gradient(self):
        model = LinearRegression()
        reconstructed_gradient = LinearRegression()
        reconstructed_gradient.a.data = torch.FloatTensor([0.5])
        reconstructed_gradient.b.data = torch.FloatTensor([1.0])
        FLModelParamUtils.set_gradient(
            model=model, reference_gradient=reconstructed_gradient
        )
        self.assertEqual(model.a.grad, reconstructed_gradient.a)
        self.assertEqual(model.b.grad, reconstructed_gradient.b)

    def test_get_mismatched_param(self):
        a_val, b_val = 0.5, 1.0

        class MismatchingLinearRegression(nn.Module):
            def __init__(self):
                super().__init__()
                self.a = nn.Parameter(torch.FloatTensor([a_val]))
                self.c = nn.Parameter(torch.FloatTensor([b_val]))

            def forward(self, x):
                return self.a + self.c * x

        model_1, model_2 = LinearRegression(), LinearRegression()
        model_1.a.data, model_1.b.data = (
            torch.FloatTensor([a_val]),
            torch.FloatTensor([b_val]),
        )
        model_2.a.data, model_2.b.data = (
            torch.FloatTensor([a_val]),
            torch.FloatTensor([b_val]),
        )

        # 1) models have same params => return an empty string
        self.assertEqual(FLModelParamUtils.get_mismatched_param([model_1, model_2]), "")

        # 2) only param 'a' is different => return 'a'
        model_2.a.data = torch.FloatTensor([b_val])
        self.assertEqual(
            FLModelParamUtils.get_mismatched_param([model_1, model_2]), "a"
        )

        # 3) only param 'b' is different => return 'b'
        model_2.a.data, model_2.b.data = (
            torch.FloatTensor([a_val]),
            torch.FloatTensor([a_val]),
        )
        self.assertEqual(
            FLModelParamUtils.get_mismatched_param([model_1, model_2]), "b"
        )

        # 4) both param 'a' and 'b' are different
        # => return the first mismatched param, which is 'a'
        model_2.a.data = torch.FloatTensor([b_val])
        self.assertEqual(
            FLModelParamUtils.get_mismatched_param([model_1, model_2]), "a"
        )

        # 5) param 'b' in model_1 is missing in MismatchingLinearRegression
        # => return 'b'
        self.assertEqual(
            FLModelParamUtils.get_mismatched_param(
                [model_1, MismatchingLinearRegression()]
            ),
            "b",
        )

    def test_copy_models(self):
        torch.manual_seed(1)
        fc_model = FCModel()
        torch.manual_seed(2)
        copied_fc_model = FCModel()
        self.assertFalse(
            FLTestUtils.do_two_models_have_same_weights(fc_model, copied_fc_model)
        )

        FLModelParamUtils.copy_models(fc_model, [copied_fc_model])
        self.assertTrue(
            FLTestUtils.do_two_models_have_same_weights(fc_model, copied_fc_model)
        )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import torch
from flsim.utils.timing.training_duration_distribution import (
    PerUserUniformDurationDistribution,
    PerUserUniformDurationDistributionConfig,
    PerUserHalfNormalDurationDistribution,
    PerUserHalfNormalDurationDistributionConfig,
    DurationDistributionFromListConfig,
    DurationDistributionFromList,
    DurationInfo,
)
from flsim.utils.timing.training_time_estimator import (
    get_training_time,
    AsyncTrainingTimeEstimator,
    SyncTrainingTimeEstimator,
)
from libfb.py import testutil
from omegaconf import OmegaConf


class TrainingTimeEstimatorTest(testutil.BaseFacebookTestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_time_from_list(self) -> None:
        """
        Test training time from list

        Assuming UPR = 2

        Sync would be the sum of slowest user between rounds
            round 1
            user_1: duration = 4
            user_2: duration = 3

            round 2
            user_3: duration = 2
            user_4: duration = 1

            total = 4 + 2 = 6

        Async would be the

            user_1: duration = 4, start_time = 1
            user_2: duration = 3, start_time = 1

            user_3: duration = 2, start_time = 2
            user_4: duration = 1, start_time = 3

            users training @ time 1: user 1, user 2
            users training @ time 3: user 2, user 3
            users training @ time 4: user 3, user 4
            users training @ time 5: user 4 finishes training
        """
        training_events = [
            DurationInfo(duration=4),
            DurationInfo(duration=3),
            DurationInfo(duration=2),
            DurationInfo(duration=1),
        ]
        async_start_times = [1, 1, 2, 3]
        sync_training_dist = DurationDistributionFromList(
            **OmegaConf.structured(
                DurationDistributionFromListConfig(training_events=training_events)
            )
        )
        async_training_dist = DurationDistributionFromList(
            **OmegaConf.structured(
                DurationDistributionFromListConfig(training_events=training_events)
            )
        )
        num_users = len(training_events)
        epochs = 1
        users_per_round = 2

        sync_estimator = SyncTrainingTimeEstimator(
            total_users=len(training_events),
            users_per_round=users_per_round,
            epochs=epochs,
            training_dist=sync_training_dist,
        )

        async_estimator = AsyncTrainingTimeEstimator(
            total_users=num_users,
            users_per_round=users_per_round,
            epochs=epochs,
            training_dist=async_training_dist,
            start_times=async_start_times,
        )

        async_time = async_estimator.training_time()
        sync_time = sync_estimator.training_time()
        self.assertEqual(sync_time, 6)
        self.assertEqual(async_time, 5)

    def test_uniform_training_time(self) -> None:
        """
        Test uniform training time

        Sync and Async should have the same training time if
        UPR = 1 and duration_min close to duration_mean
        """
        torch.manual_seed(0)
        num_users = 1000
        epochs = 1
        users_per_round = 1
        duration_mean = 1.00
        duration_min = 0.99999

        training_dist = PerUserUniformDurationDistribution(
            **OmegaConf.structured(
                PerUserUniformDurationDistributionConfig(
                    training_duration_mean=duration_mean,
                    training_duration_min=duration_min,
                )
            )
        )

        sync_time, async_time = get_training_time(
            num_users=num_users,
            users_per_round=users_per_round,
            epochs=epochs,
            training_dist=training_dist,
        )
        self.assertAlmostEqual(sync_time, async_time, delta=1e-3)

    def test_per_user_half_normal(self) -> None:
        """
        Test half normal training time

        Sync and Async should have the following training time
            sync_training_time = async_training_time = num_users * duration_min
        if UPR = 1 and duraton_std is close to 0
        """
        torch.manual_seed(0)
        num_users = 1000
        epochs = 1
        users_per_round = 1
        duration_std = 1e-6
        duration_min = 1.0

        training_dist = PerUserHalfNormalDurationDistribution(
            **OmegaConf.structured(
                PerUserHalfNormalDurationDistributionConfig(
                    training_duration_sd=duration_std,
                    training_duration_min=duration_min,
                )
            )
        )

        sync_time, async_time = get_training_time(
            num_users=num_users,
            users_per_round=users_per_round,
            epochs=epochs,
            training_dist=training_dist,
        )
        self.assertAlmostEqual(sync_time, async_time, delta=1e-3)
        self.assertAlmostEqual(sync_time, num_users * duration_min, delta=1e-3)
        self.assertAlmostEqual(async_time, num_users * duration_min, delta=1e-3)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import copy
from typing import Any, List, Optional

import torch
import torch.nn as nn
from flsim.data.data_provider import IFLDataProvider
from flsim.interfaces.metrics_reporter import IFLMetricsReporter
from flsim.interfaces.model import IFLModel
from flsim.utils.fl.common import FLModelParamUtils
from opacus import PrivacyEngine


class FLTestUtils:
    @classmethod
    def compare_model_linear_comb(cls, model1: nn.Module, model2: nn.Module):
        temp_modelA = copy.deepcopy(model1)
        temp_modelB = copy.deepcopy(model1)
        temp_modelC = copy.deepcopy(model1)
        # model1 + 0*model2 = model1
        FLModelParamUtils.linear_comb_models(model1, 1, model2, 0, temp_modelA)
        assert FLModelParamUtils.get_mismatched_param([model1, temp_modelA]) == ""
        # model1 + model2 != model1
        FLModelParamUtils.linear_comb_models(model1, 1, model2, 1, temp_modelA)
        assert FLModelParamUtils.get_mismatched_param([model1, temp_modelA]) != ""
        # (2*model1 + 3*model1 ) - 4*model1 = model1
        FLModelParamUtils.linear_comb_models(model1, 2, model1, 3, temp_modelA)
        FLModelParamUtils.linear_comb_models(model1, 4, model1, 0, temp_modelB)
        FLModelParamUtils.linear_comb_models(
            temp_modelA, 1, temp_modelB, -1, temp_modelC
        )
        assert FLModelParamUtils.get_mismatched_param([model1, temp_modelC], 1e-5) == ""
        # test that resuing one of the input models as model_to_save also works
        # model1 = model1 - model2, followed by model2 = model1 + model2
        # model2 should be the same as original model1
        temp_modelA = copy.deepcopy(model1)
        FLModelParamUtils.linear_comb_models(model1, 1, model2, -1, model1)
        FLModelParamUtils.linear_comb_models(model1, 1, model2, 1, model1)
        assert FLModelParamUtils.get_mismatched_param([model1, temp_modelA], 1e-5) == ""

    @classmethod
    def random_grad(cls, model: nn.Module):
        for param in model.parameters():
            param.grad = torch.rand_like(param)

    @classmethod
    def compare_gradient_reconstruction(
        cls, model0: nn.Module, copy_model0: nn.Module, reconstructed_grad: nn.Module
    ):
        """Test that gradient reconstruction post-optimization works
        Moment-based optimizers for FL require approximate gradient reconstruction from
        two models: original model, and new model after FL optmization step
        approx_gradient = original_model - new_model
        This test checks that gradient reconstruction works as expected
        """
        # copy model.0 into copy_model.0
        # create optimizerA for model.0, take 1 step of gradient descent on model.0,
        # moving to model.1
        # reconstruct original gradients by reconstructred_grad = model.1 - model.0
        # set grad(copy_model.0) = reconstructed_grad
        # create optimizerB for copy_model.0
        # take 1 step of gradient descent on copy_model.0, moving to copy_model.1
        # check model.1 = copy_model.1
        learning_rate = 1.0
        FLModelParamUtils.copy_models(model0, [copy_model0])
        optimizer = torch.optim.SGD(model0.parameters(), lr=learning_rate)
        # take a few steps of gradient descent
        for _i in range(0, 10):
            optimizer.zero_grad()
            cls.random_grad(model0)
            optimizer.step()

        copy_optimizer = torch.optim.SGD(copy_model0.parameters(), lr=learning_rate)
        copy_optimizer.zero_grad()
        FLModelParamUtils.reconstruct_gradient(
            old_model=copy_model0, new_model=model0, grads=reconstructed_grad
        )
        FLModelParamUtils.set_gradient(
            model=copy_model0, reference_gradient=reconstructed_grad
        )
        copy_optimizer.step()
        assert (
            FLModelParamUtils.get_mismatched_param(
                [model0, copy_model0], rel_epsilon=1e-4
            )
            == ""
        )

    @classmethod
    def _verify_averaged_and_orig_models(
        cls, orig_models: List[nn.Module], new_models: List[nn.Module]
    ) -> None:
        """Verify that:
        a) Every model in new_models is the same
        b) Every model in new_models is the 'average' of models in orig_models
        """
        assert len(orig_models) == len(new_models)
        if len(orig_models) == 0:
            return
        orig_dicts = [dict(aModel.named_parameters()) for aModel in orig_models]
        new_dicts = [dict(aModel.named_parameters()) for aModel in new_models]
        assert len(orig_dicts) == len(new_dicts)
        if len(orig_dicts) == 0:
            return
        # verify new models have all params same
        assert FLModelParamUtils.get_mismatched_param(new_models) == ""

        # verify that new_models have average of old models
        for name, param in new_dicts[0].items():
            orig_tensors = torch.stack([thedict[name] for thedict in orig_dicts])
            orig_shape = orig_tensors[0].shape
            averaged = torch.mean(orig_tensors, 0, keepdim=True)
            averaged_reshaped = averaged.view(orig_shape)
            assert torch.allclose(averaged_reshaped, param, atol=1e-6)

    @classmethod
    def average_and_verify_models(cls, orig_models: List[nn.Module]) -> None:
        """Compute the average of models in orig_models, and verify the average"""
        if len(orig_models) == 0:
            return
        models = copy.deepcopy(orig_models)
        temp_model = copy.deepcopy(models[0])
        FLModelParamUtils.average_models(models, temp_model)
        FLModelParamUtils.copy_models(temp_model, models)
        cls._verify_averaged_and_orig_models(orig_models, models)

    @classmethod
    def do_two_models_have_same_weights(cls, model1, model2) -> bool:
        for p1, p2 in zip(model1.parameters(), model2.parameters()):
            if p1.data.ne(p2.data).sum() > 0:
                return False
        return True

    @classmethod
    def train_non_fl(
        cls,
        data_provider: IFLDataProvider,
        global_model: IFLModel,
        optimizer: torch.optim.Optimizer,
        metrics_reporter: Optional[IFLMetricsReporter] = None,
        private_engine: Optional[PrivacyEngine] = None,
        epochs: int = 1,
        cuda_enabled: bool = False,
    ):
        if cuda_enabled:
            global_model.fl_cuda()
        if private_engine is not None:
            private_engine.attach(optimizer)

        for _ in range(epochs):
            for one_user_data in data_provider.train_data():
                for batch in one_user_data:
                    optimizer.zero_grad()
                    batch_metrics = global_model.fl_forward(batch)

                    if metrics_reporter is not None:
                        metrics_reporter.add_batch_metrics(batch_metrics)

                    batch_metrics.loss.backward()
                    optimizer.step()
        return global_model, metrics_reporter

    @classmethod
    def run_nonfl_training(
        cls,
        model: IFLModel,
        optimizer: torch.optim.Optimizer,
        data_loader: torch.utils.data.DataLoader,
        epochs: int,
    ) -> IFLModel:
        torch.manual_seed(1)
        for _ in range(epochs):
            for training_batch in data_loader:
                FLTestUtils.run_nonfl_training_one_batch(
                    model=model, optimizer=optimizer, training_batch=training_batch
                )
        return model

    @classmethod
    def run_nonfl_training_one_batch(
        cls, model: IFLModel, optimizer: torch.optim.Optimizer, training_batch: Any
    ):
        optimizer.zero_grad()
        batch_metrics = model.fl_forward(training_batch)
        loss = batch_metrics.loss
        loss.backward()
        optimizer.step()



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from abc import ABC, abstractmethod
from typing import Any, Iterable, Iterator, List

from flsim.interfaces.model import IFLModel


class IFLData(ABC):
    """
    Wraps data for a single entity.
    """

    @abstractmethod
    def __iter__(self) -> Iterator[Any]:
        """
        Iterator to return a user batch data
        """

    @abstractmethod
    def num_examples(self) -> int:
        """
        Returns the number of examples
        """

    @abstractmethod
    def num_batches(self) -> int:
        """
        Returns the number of batches
        """


class IFLUserData(IFLData):
    """
    Wraps data for a single user

    IFLUserData is responsible for
        1. Keeping track of the number of examples for a particular user
        2. Keeping track of the number of batches for a particular user
        3. Providing an iterator over all the user batches
    """


class IFLDataProvider(ABC):
    """
    Provides data to the trainer

    IFLDataProvider is resposible for
        1. Enforcing a uniform interface that trainer expects
        2. Transforming data into what IFLModel.fl_forward() is going to consume
        3. Keeping track of the sharded client data
    """

    def __iter__(self) -> Iterable[IFLUserData]:
        """
        Returns training data iterable
        """
        yield from self.train_data()

    def __getitem__(self, index) -> IFLUserData:
        """
        Returns a user from index
        """
        return self.get_user_data(index)

    @abstractmethod
    def user_ids(self) -> List[int]:
        """
        Returns a list of user ids in the data set
        """

    @abstractmethod
    def num_users(self) -> int:
        """
        Returns the number of users in train set
        """

    @abstractmethod
    def get_user_data(self, user_index: int) -> IFLUserData:
        """
        Returns user data from user_index
        """

    @abstractmethod
    def train_data(self) -> Iterable[IFLUserData]:
        """
        Returns training data iterable
        """

    @abstractmethod
    def eval_data(self) -> Iterable[Any]:
        """
        Returns evaluation data iterable
        """

    @abstractmethod
    def test_data(self) -> Iterable[Any]:
        """
        Returns test data iterable
        """


class FLUserDataFromList(IFLUserData):
    """
    Util class to create an IFLUserData from a list of user batches
    """

    def __init__(self, data: Iterable, model: IFLModel):
        self.data = data
        self._num_examples: int = 0
        self._num_batches: int = 0
        self.model = model
        for batch in self.data:
            training_batch = self.model.fl_create_training_batch(batch=batch)
            self._num_examples += model.get_num_examples(training_batch)
            self._num_batches += 1

    def __iter__(self):
        for batch in self.data:
            yield self.model.fl_create_training_batch(batch=batch)

    def num_batches(self):
        return self._num_batches

    def num_examples(self):
        return self._num_examples


class FLDataProviderFromList(IFLDataProvider):
    r"""
    Util class to help ease the transition to IFLDataProvider

    =======
    Args:
        train_user_list: (Iterable[Iterable[Any]]): train data
        eval_user_list: (Iterable[Any]): eval data
        test_user_list (Iterable[Any]): test data
        model: (IFLModel): the IFLModel to create training batch for
    """

    def __init__(
        self,
        train_user_list: Iterable[Iterable[Any]],
        eval_user_list: Iterable[Any],
        test_user_list: Iterable[Any],
        model: IFLModel,
    ):
        self.train_user_list = train_user_list
        self.eval_user_list = eval_user_list
        self.test_user_list = test_user_list
        self.model = model
        self._users = {
            user_id: FLUserDataFromList(user_data, model)
            for user_id, user_data in enumerate(train_user_list)
        }

    def user_ids(self):
        return list(self._users.keys())

    def num_users(self):
        return len(self.user_ids())

    def get_user_data(self, user_index: int):
        if user_index in self._users:
            return self._users[user_index]
        else:
            raise IndexError(f"Index {user_index} not in {self.user_ids()}")

    def train_data(self):
        return list(self._users.values())

    def eval_data(self):
        return [
            self.model.fl_create_training_batch(batch=batch)
            for batch in self.eval_user_list
        ]

    def test_data(self):
        return [
            self.model.fl_create_training_batch(batch=batch)
            for batch in self.test_user_list
        ]



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import Optional

from flsim.channels.base_channel import IdentityChannel
from flsim.interfaces.model import IFLModel
from flsim.reducers.dp_round_reducer import DPRoundReducer, DPRoundReducerConfig
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.fl.common import FLModelParamUtils
from omegaconf import MISSING
from torch import nn


class EstimatorType(IntEnum):
    UNBIASED = 0
    BIASED = 1


class WeightedDPRoundReducer(DPRoundReducer):
    r"""
    A differentially private round reducer that allows client models
    to provide weighted updates.

    There are two different estimators supported `BIASED` and `UNBIASED`, which only
    differ when in average reduction. For sum reduction both sensitivities
    are the same and are equivalent to ``max_weight * clipping_value``.
    """

    def __init__(
        self,
        *,
        global_model: IFLModel,
        num_users_per_round: int,
        total_number_of_users: int,
        channel: Optional[IdentityChannel] = None,
        name: Optional[str] = None,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=WeightedDPRoundReducerConfig,
            **kwargs,
        )

        super().__init__(
            global_model=global_model,
            num_users_per_round=num_users_per_round,
            total_number_of_users=total_number_of_users,
            channel=channel,
            name=name,
            **kwargs,
        )
        assert self.is_weighted, "Please use DPRoundReducer for unweighted cases"

        # pyre-fixme[16]: `WeightedDPRoundReducer` has no attribute `cfg`.
        self.min_weight = self.cfg.min_weight
        self.max_weight = self.cfg.max_weight
        self.mean_weight = self.cfg.mean_weight
        self.estimator_type = self.cfg.estimator_type
        self._check_boundaries()

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _check_boundaries(self):
        """
        Checks for min, max, and mean values of config
        """
        if (
            self.min_weight < 0
            or self.max_weight < self.min_weight
            or self.mean_weight < self.min_weight
            or self.mean_weight > self.max_weight
        ):
            self.logger.error("Weight boundaries in config are not defined properly")
        if self.estimator_type == EstimatorType.UNBIASED and self.mean_weight <= 0:
            self.logger.error(
                "For unbiased sensitivity estimation mean_weight needs to be positive."
            )

    def clamp_weight(self, weight: float) -> float:
        if not self.min_weight <= weight <= self.max_weight:
            self.logger.error(
                "min/max client weight boundaries are violated!"
                "Client weight is being adjusted"
            )
            weight = max(min(weight, self.max_weight), self.min_weight)
        return weight

    def update_reduced_module(self, delta_module: nn.Module, weight: float) -> None:
        weight = self.clamp_weight(weight)
        super().update_reduced_module(delta_module, weight)

    def check_total_weight(self, total_weight: float):
        r"""
        Boundary check for total weights.
        """
        lower_bound = self.num_users_per_round * self.min_weight
        upper_bound = self.num_users_per_round * self.max_weight
        is_bounded = lower_bound <= total_weight <= upper_bound
        if not is_bounded:
            self.logger.error(
                f"Summed weights {total_weight} do not fall within expected range [{lower_bound}, {upper_bound}]"
            )

    def sensitivity(self, total_weight: float):
        r"""
        Calculates the sensitivity of the final result.

        Note:
        Sensitivity for weighted averaging may modify the result to decrease
        sensitivity for BIASED case.
        """
        self.check_total_weight(total_weight)
        if not self.is_averaged:
            return self._sum_estimator()
        elif self.estimator_type == EstimatorType.UNBIASED:
            return self._unbiased_estimator()
        else:
            return self._biased_estimator(total_weight=total_weight)

    def _sum_estimator(self) -> float:
        return self.clipping_value * self.max_weight

    def _unbiased_estimator(self) -> float:
        """
        For weighted averaged reductions, the unbiased estimator calculates the true
        weighted average of the models and the sensitivity of it will be:
            (clipping_value * max_weight) / (min_weight * users_per_round)
        """
        return (
            (self.clipping_value * self.max_weight)
            / self.num_users_per_round
            * self.min_weight
        )

    def _biased_estimator(self, total_weight: float) -> float:
        """
        For the biased estimator the weighted average is biased, where the
        average is calculated by weighted sum of the models divided by
        max(num_clients_per_round * mean_weight, total_weight) and
        Sensitivity
        (clipping_value * max_weight) / (mean_weight * num_clients_per_round)
        """
        weight_modifier = total_weight / max(
            total_weight, self.mean_weight * self.num_users_per_round
        )
        FLModelParamUtils.linear_comb_models(
            self.reduced_module,
            weight_modifier,
            self.reduced_module,
            0.0,
            self.reduced_module,
            # pyre-fixme[16]: `WeightedDPRoundReducer` has no attribute `cfg`.
            only_federated_params=self.cfg.only_federated_params,
        )
        return (
            self.clipping_value
            * self.max_weight
            / (self.num_users_per_round * self.mean_weight)
        )


@dataclass
class WeightedDPRoundReducerConfig(DPRoundReducerConfig):
    r"""
    Contains configurations for a private round reducer based that
    also allows for weights.

    Note:
    Allowing weights in dp should generally be avoided unless weights
    are in the same range. If weights are extremely different, one
    might as well throw updates from clients with smaller weights
    away as they will be drawned in noise.
    """
    _target_: str = fullclassname(WeightedDPRoundReducer)
    min_weight: float = 1e-6
    max_weight: float = float("inf")
    mean_weight: float = 1e-6
    estimator_type: EstimatorType = EstimatorType.UNBIASED


@dataclass
class DeltaDirectionWeightedDPReducerConfig(WeightedDPRoundReducerConfig):
    _target_: str = MISSING



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import math

import torch
from sklearn.cluster import KMeans


class PQ:
    """
    Quantizes the layer weights W with the standard Product Quantization
    technique. This learns n_dict codebooks of codewords of size
    block_size from W.

    For further reference on using PQ to quantize neural networks, see
    "And the Bit Goes Down: Revisiting the Quantization of Neural Networks",
    ICLR 2020.

    PQ is performed in two steps:
    (1) The matrix W (weights or fully-connected or convolutional layer)
        is reshaped to (block_size, -1).
            - If W is fully-connected (2D), its rows are split into
              blocks of size block_size.
            - If W is convolutional (4D), its filters are split along the
              spatial dimension.
    (2) We apply the standard EM/k-means algorithm to the resulting reshaped matrix.

    Once W is reshaped to (block_size, n_samples), we learn num_codebooks codebooks
    each of size n_samples // num_codebooks (except the last which may have a variable
    size). More specifically, the first num_codebooks samples belong to the first dict,
    and so on. We use a trick to recover the quantized matrix from the knowledge of
    its centroids and assignments: we shift the assignments by a factor n_centroids
    every dict. See the decode() function.

    Args:
        - sizes: sizes of the weight matrix to quantize
        - max_block_size: max allowed block size (for the subvectors)
        - num_codebooks: number of dicts
        - max_num_centroids: max allowed number of centroids
        - num_k_means_iter: number of k-means iterations
        - verbose: print information after each iteration

    Notes:
        - PQ works for tensors that are on the CPU or on the GPU.
        - We need the original size of the weight matrix to decode, that's why
          we include it in the class state.
        - We compute internally the actual block_size in _determine_block_size.
          The actual block size is defined as the largest block size that is
          compatible with the shape of W while being less or equal than max_block_size.
        - We compute internally the actual number of centroids in _determine_num_centroids
          to avoid quantizing small layers with too much centroids.
    """

    def __init__(
        self,
        sizes: torch.Size,
        max_block_size: int = 9,
        num_codebooks: int = 1,
        max_num_centroids: int = 256,
        num_k_means_iter: int = 20,
        verbose: bool = False,
    ):
        self.sizes = sizes
        self.ndim = len(sizes)
        self.num_codebooks = num_codebooks
        self.num_k_means_iter = num_k_means_iter
        self.verbose = verbose
        self.block_size = self._determine_block_size(max_block_size)
        self.n_centroids = self._determine_num_centroids(max_num_centroids)

    def _determine_block_size(self, max_block_size):
        """
        Return the largest block size that is compatible with
        the shape of W while being less than or equal to max_block_size.
        """

        if self.ndim == 2:
            _out_features, in_features = self.sizes
            allowed_block_sizes = filter(
                lambda block_size: in_features % block_size == 0,
                range(1, max_block_size + 1),
            )
            block_size = list(allowed_block_sizes)[-1]

        elif self.ndim == 4:
            _out_channels, in_channels, kh, kw = self.sizes
            allowed_block_sizes = filter(
                lambda block_size: (in_channels * kh * kw) % block_size == 0,
                range(1, max_block_size + 1),
            )
            block_size = list(allowed_block_sizes)[-1]

        else:
            raise NotImplementedError(self.sizes)

        if self.verbose:
            print(f"Selected block size {block_size} for W of shape {self.sizes}")

        return block_size

    def _determine_num_centroids(self, max_num_centroids, max_centroid_factor_bound=4):
        """
        W is split into n_subvectors per dict. Returns n_centroids such that:
            - n_centroids is a power of two (greater or equal than 2)
            - n_centroids <= max_num_centroids
            - n_centroids * max_centroid_factor_bound < n_subvectors

        Notes:
            - This is to avoid quantizing small layers with too much centroids.
            - Must be called after determining self.block_size.
        """

        n_tot_subvectors = math.prod(self.sizes) // self.block_size
        n_subvectors = n_tot_subvectors // self.num_codebooks
        assert n_subvectors >= 8, "Not enough subvectors, consider not quantizing."
        n_centroids = 2 ** int(math.log2(n_subvectors // max_centroid_factor_bound))
        n_centroids = min(max_num_centroids, n_centroids)

        if self.verbose:
            print(f"Selected n_centroids {n_centroids} for W of shape {self.sizes}")

        return n_centroids

    def _reshape_and_split(self, W) -> torch.Tensor:
        """
        Reshapes the matrix W as expained in step (1).
        """

        # fully connected: by convention the weight has size out_features x in_features
        if self.ndim == 2:
            out_features, in_features = self.sizes
            assert (
                in_features % self.block_size == 0
            ), "Linear: in_features must be a multiple of block_size"
            W_unsplit = (
                W.reshape(out_features, -1, self.block_size)
                .permute(2, 1, 0)
                .flatten(1, 2)
            )

        # convolutional: we reshape along the spatial dimension
        elif self.ndim == 4:
            out_channels, in_channels, kh, kw = self.sizes
            assert (
                in_channels * kh * kw
            ) % self.block_size == 0, (
                "Conv: kernel_size kh * kw must be a multiple of block_size"
            )
            W_unsplit = (
                W.reshape(out_channels, -1, self.block_size)
                .permute(2, 0, 1)
                .flatten(1, 2)
            )

        # not implemented
        else:
            raise NotImplementedError(self.sizes)

        # split into self.num_codebooks blocks (last block may be larger)
        split = W_unsplit.size(1) // self.num_codebooks
        last_split = W_unsplit.size(1) - split * (self.num_codebooks - 1)
        splits = [split] * (self.num_codebooks - 1) + [last_split]
        return torch.split(W_unsplit, splits, dim=1)

    def _offset_assignments(self, assignments: torch.Tensor) -> torch.Tensor:
        """
        See ``decode`` for an explanation and illustration.
        """

        n_assignments = len(assignments)
        subvectors_per_dict = int(math.ceil(n_assignments / self.num_codebooks))
        offset = torch.arange(
            0, self.num_codebooks * self.n_centroids, self.n_centroids
        )
        offset = offset.type_as(assignments)
        offset = offset.repeat_interleave(subvectors_per_dict)[:n_assignments]
        return assignments + offset

    def encode(self, W):
        """
        Performs num_k_means_iter EM steps as explained in step (2).
        """

        # reshape and split W as expained in step (1).
        W_reshaped = self._reshape_and_split(W)
        # compute centroids for all dicts
        all_centroids = []
        all_assignments = []
        for d in range(self.num_codebooks):
            if self.verbose:
                print(
                    f"Building dict {d+1}/{self.num_codebooks} with {self.n_centroids} "
                    f"centroids for {W_reshaped[d].size(1)} vectors"
                )

            # current weight
            W_curr = W_reshaped[d]

            # run k-means
            kmeans = KMeans(
                n_clusters=self.n_centroids,
                init="k-means++",
                n_init=1,
                max_iter=self.num_k_means_iter,
                tol=0.0001,
                verbose=self.verbose,
            )
            assignments = kmeans.fit_predict(W_curr.t())
            centroids = kmeans.cluster_centers_
            assignments = torch.LongTensor(assignments)
            centroids = torch.Tensor(centroids)

            # remember centroids and assignments
            all_centroids.append(centroids)
            all_assignments.append(assignments)

        # cat centroids and assignments
        assignments = torch.cat(all_assignments)
        assignments = self._offset_assignments(assignments)
        centroids = torch.cat(all_centroids)
        return centroids, assignments

    def decode(
        self,
        centroids: torch.Tensor,
        assignments: torch.Tensor,
    ) -> torch.Tensor:
        """
        Returns the encoded full weight matrix. Must be called after
        the encode function.

        We offset assignments, let us illustrate this on an example.
        Say num_codebooks = 2 with 3 centroids per dict, and assume that
        assignments = [1, 2, 3, 3, 1, 1, 3, 2]. Then, after the offset
        the assignments would be [1, 2, 3, 3, 4, 4, 6, 5].

        Thus, we can call centroids[assignments] to properly recover W.

        Args:
            - centroids has size (num_codebooks x n_centroids, block_size)
            - assignments has size (n_samples)
        """

        # decode in the fully connected case
        if self.ndim == 2:
            out_features, _ = self.sizes
            return (
                centroids[assignments]
                .reshape(-1, out_features, self.block_size)
                .permute(1, 0, 2)
                .flatten(1, 2)
            )

        # decode in the convolutional case
        elif self.ndim == 4:
            out_channels, in_channels, kh, kw = self.sizes
            return (
                centroids[assignments]
                .reshape(-1, in_channels, self.block_size)
                .permute(0, 1, 2)
                .reshape(out_channels, in_channels, kh, kw)
            )

        # not implemented
        else:
            raise NotImplementedError(
                f"Only supports 2D convolutions and linear layers, but got size {self.sizes}"
            )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from typing import Any, Dict, Union

from flsim.optimizers.async_aggregators import (
    FedAdamAsyncAggregatorConfig,
    FedAvgWithLRAsyncAggregatorConfig,
    FedAdamHybridAggregatorConfig,
    FedAvgWithLRHybridAggregatorConfig,
    AsyncAggregator,
    FedAdamAsyncAggregator,
    FedAvgWithLRAsyncAggregator,
)
from flsim.optimizers.sync_aggregators import (
    FedAdamSyncAggregatorConfig,
    FedAvgWithLRSyncAggregatorConfig,
    FedAvgSyncAggregatorConfig,
    FedAdamSyncAggregator,
    FedAvgWithLRSyncAggregator,
    FedAvgSyncAggregator,
    SyncAggregator,
)
from flsim.trainers.async_trainer import AsyncTrainer
from flsim.trainers.sync_trainer import SyncTrainer
from omegaconf import OmegaConf
from torch.optim.optimizer import Optimizer


class OptimizerTestUtil:
    SYNC_AGGREGATOR_TEST_CONFIGS = {
        "avg": {"dict_config": {"_base_": "base_fed_avg_sync_aggregator"}},
        "sgd": {
            "dict_config": {
                "_base_": "base_fed_avg_with_lr_sync_aggregator",
                "lr": 0.1,
                "momentum": 0.9,
            }
        },
        "adam": {
            "dict_config": {
                "_base_": "base_fed_adam_sync_aggregator",
                "lr": 0.1,
                "weight_decay": 0.9,
                "eps": 1e-8,
            }
        },
    }

    ASYNC_AGGREGATOR_TEST_CONFIGS = {
        "sgd": {
            "dict_config": {
                "_base_": "base_fed_avg_with_lr_async_aggregator",
                "lr": 0.1,
                "momentum": 0.9,
            }
        },
        "adam": {
            "dict_config": {
                "_base_": "base_fed_adam_async_aggregator",
                "lr": 0.1,
                "weight_decay": 0.9,
                "eps": 1e-8,
            }
        },
    }

    @classmethod
    def provide_sync_factory_creation_dataset(cls):
        return [
            (
                "FedAvg",
                OmegaConf.structured(FedAvgSyncAggregatorConfig),
            ),
            (
                "FedAvgWithLR",
                OmegaConf.structured(
                    FedAvgWithLRSyncAggregatorConfig(lr=0.1, momentum=0.9)
                ),
            ),
            (
                "FedAdam",
                OmegaConf.structured(
                    FedAdamSyncAggregatorConfig(lr=0.1, weight_decay=0.9, eps=1e-8)
                ),
            ),
        ]

    @classmethod
    def provide_async_factory_creation_dataset(cls):
        return [
            (
                "FedAvgWithLR",
                OmegaConf.structured(
                    FedAvgWithLRAsyncAggregatorConfig(lr=0.1, momentum=0.9)
                ),
            ),
            (
                "FedAdam",
                OmegaConf.structured(
                    FedAdamAsyncAggregatorConfig(lr=0.1, weight_decay=0.9, eps=1e-8)
                ),
            ),
        ]

    @classmethod
    def provide_hybrid_factory_creation_dataset(cls):
        return [
            (
                "FedAvgWithLR",
                OmegaConf.structured(
                    FedAvgWithLRHybridAggregatorConfig(
                        lr=0.1, momentum=0.9, buffer_size=3
                    )
                ),
            ),
            (
                "FedAdam",
                OmegaConf.structured(
                    FedAdamHybridAggregatorConfig(
                        lr=0.1,
                        weight_decay=0.9,
                        eps=1e-8,
                        buffer_size=3,
                    )
                ),
            ),
        ]

    @classmethod
    def get_value_from_optimizer(cls, optimizer: Optimizer, attribute_name: str):
        for param_group in optimizer.param_groups:
            return param_group[attribute_name]
        raise TypeError(f"Optimizer does not have attribute: {attribute_name}")

    @classmethod
    def get_sync_aggregator_test_configs(cls):
        return list(cls.SYNC_AGGREGATOR_TEST_CONFIGS.values())

    @classmethod
    def get_async_aggregator_test_configs(cls):
        return list(cls.ASYNC_AGGREGATOR_TEST_CONFIGS.values())

    @classmethod
    def _verify_trainer_common_aggregators(
        cls,
        aggregator: Union[SyncAggregator, AsyncAggregator],
        dict_config: Dict[str, Any],
    ):
        if "fed_avg_with_lr" in dict_config["_base_"]:
            assert isinstance(aggregator, FedAvgWithLRSyncAggregator) or isinstance(
                aggregator, FedAvgWithLRAsyncAggregator
            )
            assert (
                OptimizerTestUtil.get_value_from_optimizer(aggregator.optimizer, "lr")
                == 0.1
            ), "lr for FedavgwithLr optimizer should be 0.1"
            assert (
                OptimizerTestUtil.get_value_from_optimizer(
                    aggregator.optimizer, "momentum"
                )
                == 0.9
            ), "momentum for FedavgwithLr optimizer should be 0.9"
        if "fed_adam" in dict_config["_base_"]:
            assert isinstance(aggregator, FedAdamSyncAggregator) or isinstance(
                aggregator, FedAdamAsyncAggregator
            )
            assert (
                OptimizerTestUtil.get_value_from_optimizer(aggregator.optimizer, "lr")
                == 0.1
            ), "lr for adam optimizer should be 0.1"
            assert (
                OptimizerTestUtil.get_value_from_optimizer(
                    aggregator.optimizer, "weight_decay"
                )
                == 0.9
            ), "weight_decay for adam optimizer should be 0.9"
            assert (
                OptimizerTestUtil.get_value_from_optimizer(aggregator.optimizer, "eps")
                == 1e-8
            ), "eps for adam optimizer should be 1e-8"

    @classmethod
    def verify_async_trainer_aggregator(
        cls, trainer: AsyncTrainer, dict_config: Dict[str, Any]
    ):
        aggregator = trainer.aggregator
        assert isinstance(
            aggregator, AsyncAggregator
        ), "aggregator should be an instance of AsyncAggregator for AsyncTrainer"
        cls._verify_trainer_common_aggregators(aggregator, dict_config)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import copy
from typing import Tuple

import numpy as np
import torch
from flsim.common.pytest_helper import (
    assertEqual,
    assertEmpty,
)
from flsim.data.data_provider import FLDataProviderFromList, IFLDataProvider
from flsim.data.data_sharder import RoundRobinSharder
from flsim.data.dataset_data_loader import FLDatasetDataLoaderWithBatch
from flsim.interfaces.model import IFLModel
from flsim.optimizers.async_aggregators import (
    FedAvgWithLRAsyncAggregatorConfig,
    FedAvgWithLRHybridAggregatorConfig,
)
from flsim.tests.utils import verify_models_equivalent_after_training
from flsim.utils.async_trainer.async_example_weights import (
    AsyncExampleWeightConfig,
    ExampleWeight,
    EqualExampleWeightConfig,
    LinearExampleWeightConfig,
    SqrtExampleWeightConfig,
    Log10ExampleWeightConfig,
)
from flsim.utils.async_trainer.async_staleness_weights import (
    ConstantStalenessWeightConfig,
    PolynomialStalenessWeightConfig,
    ThresholdStalenessWeightConfig,
)
from flsim.utils.async_trainer.training_event_generator import (
    AsyncTrainingEventGeneratorFromListConfig,
    EventTimingInfo,
)
from flsim.utils.sample_model import DummyAlphabetFLModel
from flsim.utils.tests.helpers.async_trainer_test_utils import (
    async_train_one_user,
    get_equal_split_data,
    get_fl_data_provider,
    run_fl_nonfl_training,
    run_fl_training,
    run_fl_training_with_event_generator,
)
from flsim.utils.tests.helpers.test_data_utils import DummyAlphabetDataset
from hydra.utils import instantiate


class TestAsyncTrainerWeights:
    def _get_fl_local_lr_and_example_weights_config(
        self,
        example_weight_config: AsyncExampleWeightConfig,
        num_examples_per_user: int,
        normalize_wts: bool,
    ) -> Tuple[float, AsyncExampleWeightConfig]:
        """Compute the fl_local_learning_rate and example_weight_config
        that will cause FL and Non-FL training to produce the same results.
        Normalize weights: when this option is true, example weight is divided
        by the average example weight
        """
        return_config = copy.deepcopy(example_weight_config)
        if normalize_wts:
            # effectively equal to not having any example weight, since all users
            # in the test have the same number of examples
            return_config.avg_num_examples = num_examples_per_user
            return (1.0, return_config)
        else:
            # create un-normalized example weight object
            return_config.avg_num_examples = 1
            example_weight_obj: ExampleWeight = instantiate(return_config)
            example_weight = example_weight_obj.weight(num_examples_per_user)
            fl_local_lr = 1.0 / example_weight
            return (fl_local_lr, return_config)

    def compare_num_examples_wt(
        self,
        base_learning_rate: float,
        example_weight_config: AsyncExampleWeightConfig,
        rel_epsilon: float,
        normalize_wts: bool,
    ):
        """Run training for two tasks:
        data1: 2 training examples, same for data2 and data3
        example_weight_type: defines how number of examples affect weight
          Example_weight_str is used to chose a function ExWt(x)
          When example_weight_str = 'linear', ExWt(x) = x
          When example_weight_str = 'sqrt', ExWt(x) = sqrt(x)
          When example_weight_str = 'log10', ExWt(x) = log10(x)
        Task1: Non-FL task, data = {data1, data2, data3}, LR=base_lr
        Task2: FL async task, training_duration = 0,
                LR=base_lr/ExWt(2), example_weight_type=LINEAR
                data = {user1: data1, user2: data2, user3: data3}
                Use  user selector,
                so we train in order user1-->user2-->user3
                user_epochs_per_round=1
        Train Task1, Task2. Show that you end up with the same model
        """
        num_training_examples_per_user = 2
        num_users = 3
        num_examples = 6

        torch.manual_seed(1)

        global_model = DummyAlphabetFLModel()
        # will be used later to verify training indeed took place
        global_model_init_copy = copy.deepcopy(global_model)
        nonfl_model = copy.deepcopy(global_model_init_copy)

        (
            fl_local_learning_rate,
            example_weight_config,
        ) = self._get_fl_local_lr_and_example_weights_config(
            example_weight_config=example_weight_config,
            num_examples_per_user=num_training_examples_per_user,
            normalize_wts=normalize_wts,
        )
        fl_aggregator_config = FedAvgWithLRAsyncAggregatorConfig(lr=base_learning_rate)
        nonfl_optimizer = torch.optim.SGD(
            nonfl_model.fl_get_module().parameters(), lr=base_learning_rate
        )
        num_epochs = 1
        fl_data_provider, nonfl_data_loader = get_equal_split_data(
            num_examples=num_examples,
            num_fl_users=num_users,
            fl_batch_size=num_training_examples_per_user,
            model=global_model,
            one_batch_per_user_only=False,
        )
        fl_trained_model, nonfl_trained_model = run_fl_nonfl_training(
            fl_model=global_model,
            nonfl_model=nonfl_model,
            nonfl_optimizer=nonfl_optimizer,
            fl_data_provider=fl_data_provider,
            nonfl_data_loader=nonfl_data_loader,
            epochs=num_epochs,
            fl_local_lr=fl_local_learning_rate,
            fl_aggregator_config=fl_aggregator_config,
            training_rate=1,
            training_duration_mean=0,
            training_duration_sd=0,
            example_weight_config=example_weight_config,
        )
        assertEqual(
            verify_models_equivalent_after_training(
                fl_trained_model,
                nonfl_trained_model,
                global_model_init_copy,
                rel_epsilon=1e-4,
                abs_epsilon=1e-6,
            ),
            "",
        )

    def test_linear_example_wts_sgd(self):
        for example_weight_config in (
            EqualExampleWeightConfig(),
            LinearExampleWeightConfig(),
            SqrtExampleWeightConfig(),
            Log10ExampleWeightConfig(),
        ):
            for normalize_wts in [True, False]:
                # choose a random learning rate between 0 and 1
                learning_rate = np.random.random_sample()
                # adding a printf to debug flaky test failure in Sandcastle
                print(f"Learning rate: {learning_rate}")
                self.compare_num_examples_wt(
                    base_learning_rate=learning_rate,
                    rel_epsilon=1e-6,
                    example_weight_config=example_weight_config,
                    normalize_wts=normalize_wts,
                )

    # TODO: add test_linear_example_wts_adam once we know how to scale
    # momentum and adaptive learning rate appropriately with weights

    def _get_fl_data_round_robin_sharding(
        self, num_examples: int, num_fl_users: int, fl_batch_size: int, model: IFLModel
    ) -> IFLDataProvider:
        dummy_dataset = DummyAlphabetDataset(num_examples)
        fl_data_sharder = RoundRobinSharder(num_shards=num_fl_users)
        fl_data_loader = FLDatasetDataLoaderWithBatch(
            dummy_dataset,
            dummy_dataset,
            dummy_dataset,
            fl_data_sharder,
            fl_batch_size,
            fl_batch_size,
            fl_batch_size,
        )
        fl_data_provider = FLDataProviderFromList(
            fl_data_loader.fl_train_set(),
            fl_data_loader.fl_eval_set(),
            fl_data_loader.fl_test_set(),
            model,
        )
        assert fl_data_loader.num_total_users == num_fl_users, "Error in data sharding"
        return fl_data_provider

    def test_default_example_weights(self):
        """Create an FL Async task with default values for example weight (equal example weight)
        Note: In async, weight of a user update = example_weight * staleness_weight
        Equal example weight => example_weight = 1.0 irrespective of #examples
        Verify that weight doesn't depend on number of examples. Train 2 FL tasks:
        Task1: {user1:data1}, {user2:data2}
        Task2: {user1:data1,data1}, {user2:data2,data2} (each user gets duplicate copy of data)
        Both tasks are trained in "sequential" order
        Task1 and Task2 produce the same model
        """
        num_epochs = 1
        num_users = 2
        torch.manual_seed(1)
        # create dummy FL model on alphabet
        trained_models = []
        initial_model = DummyAlphabetFLModel()
        local_lr = np.random.random_sample() * 10
        global_lr = np.random.random_sample() * 10
        dataset_size = 26  # number of examples in DummyAlphabetDataset
        for replication_factor in [1, 2]:
            num_examples = replication_factor * dataset_size  # either 26 or 52
            fl_model = copy.deepcopy(initial_model)
            fl_data_provider = self._get_fl_data_round_robin_sharding(
                num_examples=num_examples,
                num_fl_users=num_users,
                fl_batch_size=dataset_size,
                model=fl_model,
            )
            fl_trained_model, _ = run_fl_training(
                fl_model=fl_model,
                fl_data_provider=fl_data_provider,
                epochs=num_epochs,
                local_lr=local_lr,
                aggregator_config=FedAvgWithLRAsyncAggregatorConfig(lr=global_lr),
                training_rate=1,
                training_duration_mean=0,
                training_duration_sd=0,
            )
            trained_models.append(fl_trained_model)
        assertEqual(len(trained_models), 2, "Expected to train two models")
        assertEqual(
            verify_models_equivalent_after_training(
                trained_models[0],
                trained_models[1],
                initial_model,
                rel_epsilon=1e-4,
                abs_epsilon=1e-6,
            ),
            "",
        )

    def test_default_staleness_weights(self):
        r"""Create an FL Async task with default values for staleness weight (no staleness weight)
        Verify that weight doesn't depend on staleness. Train on two users, where training events
        are as follows. Assume initial model=M
        T0: UserA starts training. Gets initial model M.
        T1: UserB starts training. Gets initial model M.
        T2: UserA finishes training. Local training: M->M_A. Global model updated to M_A
        T3: User B finishes training. Local training: M->M_B.
            Global model update: M_A->M_Final
            M_Final = M_A + (M_B - M)*staleness. **Since staleness=1**
            M_Final = M_A + M_B - M
        """
        num_epochs = 1
        torch.manual_seed(1)
        # create dummy FL model on alphabet
        initial_model = DummyAlphabetFLModel()
        local_lr = np.random.random_sample() * 10
        global_lr = 1.0
        fl_model = copy.deepcopy(initial_model)
        num_examples = 26
        num_fl_users = 2
        fl_data_provider, nonfl_data_loader = get_equal_split_data(
            num_examples=num_examples,
            num_fl_users=num_fl_users,
            fl_batch_size=num_examples // num_fl_users,  # each user has exactly 1 batch
            model=initial_model,
            one_batch_per_user_only=False,
        )
        fl_trained_model, _ = run_fl_training(
            fl_model=fl_model,
            fl_data_provider=fl_data_provider,
            epochs=num_epochs,
            local_lr=local_lr,
            aggregator_config=FedAvgWithLRAsyncAggregatorConfig(lr=global_lr),
            training_rate=1,
            training_duration_mean=100,
            training_duration_sd=0,
        )
        # Simulate FL training: train on user1, update global model. Then train
        # on user 2, update global model.
        data_loader_iter = iter(nonfl_data_loader)
        user1_batch = next(data_loader_iter)
        user2_batch = next(data_loader_iter)
        simulated_global_model = copy.deepcopy(initial_model)
        # train user1
        simulated_global_model = async_train_one_user(
            global_model_at_training_start=simulated_global_model,
            global_model_at_training_end=simulated_global_model,
            batches=[user1_batch],
            local_lr=local_lr,
        )
        # train user2
        # key difference from training user1: global_model_at_training_start
        # is different from global_model_at_training_end
        simulated_global_model = async_train_one_user(
            global_model_at_training_start=initial_model,
            global_model_at_training_end=simulated_global_model,
            batches=[user2_batch],
            local_lr=local_lr,
        )
        error_msg = verify_models_equivalent_after_training(
            fl_trained_model,
            simulated_global_model,
            initial_model,
            rel_epsilon=1e-4,
            abs_epsilon=1e-6,
        )
        assertEqual(error_msg, "")

    def test_threshold_staleness_weights(self):
        r"""
        Run training for two tasks:
        Task1: FL async task, training_duration = 100, training_rate=1
                data = {user1: data1, user2: data2, user3: data3}
                staleness_weight = {threshold. if staleness>0, weight=0}
                timeline is:
                t=0, user 1 starts training
                t=1, user 2 strats training
                t=3, user 3 starts training
                t=100, user 1 finishes training. staleness=0. staleness_wt=1.0
                t=101, user 2 finishes training. staleness=1. staleness_wt=0
                t=102, user 3 finishes training. staleness=2. stalenes_wt=0
                Because users 2 and 3 have non-zero staleness, their model updates
                get no weight at all, so it's as if they didn't even exist
        Task2: {user1: data1}
        Train Task1, Task2. Show that we end up with the same model
        """
        num_epochs = 1
        num_users = 3
        num_examples_per_user = 8
        # training config when 3 users, with staleness
        staleness_training_rate = 1
        staleness_training_duration_mean = 100
        staleness_cutoff = 0
        staleness_wt_after_cutoff = 0.0

        # training config that forces sequential training
        sequential_training_rate = 1
        sequential_training_duration_mean = 0

        local_lr = 1.0
        aggregator_config = FedAvgWithLRAsyncAggregatorConfig(lr=1.0)

        # set seed to 1 before model init
        torch.manual_seed(1)
        initial_model = DummyAlphabetFLModel()

        fl_data_provider_with_staleness, _ = get_equal_split_data(
            num_examples=num_users * num_examples_per_user,
            num_fl_users=num_users,
            fl_batch_size=num_examples_per_user,
            model=initial_model,
            one_batch_per_user_only=False,
        )
        fl_trained_model_with_staleness, _ = run_fl_training(
            fl_model=copy.deepcopy(initial_model),
            fl_data_provider=fl_data_provider_with_staleness,
            epochs=num_epochs,
            local_lr=local_lr,
            aggregator_config=aggregator_config,
            training_rate=staleness_training_rate,
            training_duration_mean=staleness_training_duration_mean,
            training_duration_sd=0,
            staleness_weight_config=ThresholdStalenessWeightConfig(
                avg_staleness=0,
                cutoff=staleness_cutoff,
                value_after_cutoff=staleness_wt_after_cutoff,
            ),
        )

        fl_data_provider_sequential, _ = get_equal_split_data(
            num_examples=num_examples_per_user,
            num_fl_users=1,
            fl_batch_size=num_examples_per_user,
            model=initial_model,
            one_batch_per_user_only=False,
        )
        fl_trained_model_sequential, _ = run_fl_training(
            fl_model=copy.deepcopy(initial_model),
            fl_data_provider=fl_data_provider_sequential,
            epochs=num_epochs,
            local_lr=local_lr,
            aggregator_config=aggregator_config,
            training_rate=sequential_training_rate,
            training_duration_mean=sequential_training_duration_mean,
            training_duration_sd=0,
            staleness_weight_config=ConstantStalenessWeightConfig(),
        )
        assertEqual(
            verify_models_equivalent_after_training(
                fl_trained_model_with_staleness,
                fl_trained_model_sequential,
                initial_model,
                rel_epsilon=1e-4,
                abs_epsilon=1e-6,
            ),
            "",
        )

    def test_polynomial_staleness_weights(self):
        r"""
        Integration test for polynomial staleness weight
        Train task where polynomial staleness weight cancels out example_weight,
        resulting in total_wt=1.0
        training_data: {
            data1 = 26 examples
            user1:{data1, data1},
            user2:{data1,}
        Task1: FL async task, training_duration = 100, training_rate=1
                staleness_weight = {
                    polynomial, exponent=1.0 => wt = 1/(staleness+1)
                    staleness=0, wt=1.
                    staleness=1, wt=1/2,
                }
                example_weight = linear, avg_num_examples=26
                timeline is:
                t=1, user 1 starts training
                t=2, user 2 starts training
                t=3, user 2 finishes training
                       staleness=0. staleness_wt=1.0, example_wt=26/26, total_wt=1
                t=101, user 1 finishes training.
                       staleness=1. staleness_wt=0.5, example_wt=52/26, total_wt=1
        Task2: FL async task, training_duration = 100, training_rate=1
                staleness_weight = constant
                example_weight = constant
                all users get total_weight = 1.0
        Train Task1, Task2. Show that we end up with the same model
        """
        # common configs
        num_epochs = 1
        # set seed to 1 before model init
        torch.manual_seed(1)
        # train fl fask

        local_lr = 1.0
        aggregator_config = FedAvgWithLRAsyncAggregatorConfig(lr=1.0)

        num_users = 2
        # user1 has dataset_size examples, user2 has 2*dataset_size examples
        dataset_size = 26
        max_num_examples_per_user = dataset_size * num_users
        total_num_examples = dataset_size * 3

        # create config such that training events happen in the following order
        # t=1, user 1 starts training
        # t=2, user 2 starts training
        # t=3, user 2 finishes training
        # t=101, user 1 finishes training.
        user_1_start_time_delta = 1
        user_2_start_time_delta = 1
        user_1_training_duration = 100
        user_2_training_duration = 1
        user1_training_events = EventTimingInfo(
            prev_event_start_to_current_start=user_1_start_time_delta,
            duration=user_1_training_duration,
        )
        user2_training_events = EventTimingInfo(
            prev_event_start_to_current_start=user_2_start_time_delta,
            duration=user_2_training_duration,
        )
        event_generator_config = AsyncTrainingEventGeneratorFromListConfig(
            training_events=[user1_training_events, user2_training_events]
        )

        # set seed to 1 before model init
        torch.manual_seed(1)
        initial_model = DummyAlphabetFLModel()

        # split data such that
        # user1: {a,b,c...z,a,b,c...z}, user2:{a,b,c...z}
        data_provider = get_fl_data_provider(
            num_examples=total_num_examples,
            num_fl_users=2,
            examples_per_user=max_num_examples_per_user,
            batch_size=max_num_examples_per_user,
            model=initial_model,
        )

        poly_staleness_wt_model = run_fl_training_with_event_generator(
            fl_model=copy.deepcopy(initial_model),
            fl_data_provider=data_provider,
            epochs=num_epochs,
            local_lr=local_lr,
            aggregator_config=aggregator_config,
            training_event_generator_config=event_generator_config,
            staleness_weight_config=PolynomialStalenessWeightConfig(
                avg_staleness=0, exponent=1.0
            ),
            example_weight_config=LinearExampleWeightConfig(
                avg_num_examples=dataset_size
            ),
        )

        equal_staleness_wt_model = run_fl_training_with_event_generator(
            fl_model=copy.deepcopy(initial_model),
            fl_data_provider=data_provider,
            epochs=num_epochs,
            local_lr=local_lr,
            aggregator_config=aggregator_config,
            training_event_generator_config=event_generator_config,
            staleness_weight_config=ConstantStalenessWeightConfig(avg_staleness=0),
            example_weight_config=LinearExampleWeightConfig(
                avg_num_examples=dataset_size
            ),
        )

        assertEqual(
            verify_models_equivalent_after_training(
                poly_staleness_wt_model,
                equal_staleness_wt_model,
                initial_model,
                rel_epsilon=1e-4,
                abs_epsilon=1e-6,
            ),
            "",
        )

    def test_example_wts_and_staleness_wts_lr_scale(self):
        """
        Test for same number of examples and a fixed global learning rate.
        Async with no example weight should be the same as linear example weight where
        ex_local_lr= 1, ex_global_lr = 1
        no_ex_local_lr = ex_local_lr, no_ex_global_lr = ex_global_lr / max_num_examples
        """
        total_num_examples = 100
        num_users = 10
        examples_per_user = total_num_examples // num_users
        batch_size = 3

        init_model = DummyAlphabetFLModel()
        data_provider = get_fl_data_provider(
            num_examples=total_num_examples,
            num_fl_users=num_users,
            examples_per_user=examples_per_user,
            batch_size=batch_size,
            model=init_model,
        )

        local_lr = 1.0
        epochs = 5
        training_mean = 1
        training_std = 1
        buffer_size = 5

        example_wt_off_model = copy.deepcopy(init_model)
        example_wt_off_global_lr = 1.0
        example_wt_off_model, _ = run_fl_training(
            fl_model=example_wt_off_model,
            fl_data_provider=data_provider,
            epochs=epochs,
            local_lr=local_lr,
            aggregator_config=FedAvgWithLRHybridAggregatorConfig(
                lr=example_wt_off_global_lr, buffer_size=buffer_size
            ),
            training_rate=1,
            training_duration_mean=training_mean,
            training_duration_sd=training_std,
            staleness_weight_config=PolynomialStalenessWeightConfig(
                avg_staleness=0, exponent=0.5
            ),
            example_weight_config=EqualExampleWeightConfig(avg_num_examples=1),
        )

        example_wt_on_model = copy.deepcopy(init_model)
        example_wt_on_global_lr = example_wt_off_global_lr / examples_per_user
        example_wt_on_model, _ = run_fl_training(
            fl_model=example_wt_on_model,
            fl_data_provider=data_provider,
            epochs=epochs,
            local_lr=local_lr,
            aggregator_config=FedAvgWithLRHybridAggregatorConfig(
                lr=example_wt_on_global_lr, buffer_size=buffer_size
            ),
            training_rate=1,
            training_duration_mean=training_mean,
            training_duration_sd=training_std,
            staleness_weight_config=PolynomialStalenessWeightConfig(
                avg_staleness=0, exponent=0.5
            ),
            example_weight_config=LinearExampleWeightConfig(avg_num_examples=1),
        )

        error_msg = verify_models_equivalent_after_training(
            example_wt_on_model,
            example_wt_off_model,
            init_model,
            rel_epsilon=1e-4,
            abs_epsilon=1e-6,
        )
        assertEmpty(error_msg, msg=error_msg)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from hydra.core.config_store import ConfigStore

from .base_channel import FLChannelConfig
from .half_precision_channel import HalfPrecisionChannelConfig
from .product_quantization_channel import ProductQuantizationChannelConfig
from .scalar_quantization_channel import ScalarQuantizationChannelConfig
from .sketch_channel import SketchChannelConfig
from .sparse_mask_channel import SparseMaskChannelConfig

ConfigStore.instance().store(
    name="base_identity_channel",
    node=FLChannelConfig,
    group="channel",
)

ConfigStore.instance().store(
    name="base_sparse_mask_channel",
    node=SparseMaskChannelConfig,
    group="channel",
)

ConfigStore.instance().store(
    name="base_sketch_channel",
    node=SketchChannelConfig,
    group="channel",
)

ConfigStore.instance().store(
    name="base_half_precision_channel",
    node=HalfPrecisionChannelConfig,
    group="channel",
)

ConfigStore.instance().store(
    name="base_scalar_quantization_channel",
    node=ScalarQuantizationChannelConfig,
    group="channel",
)

ConfigStore.instance().store(
    name="base_product_quantization_channel",
    node=ProductQuantizationChannelConfig,
    group="channel",
)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

"""
Note: Ignore pyre errors here, because we are dynamically instantiating
class attributes for our Configs and Pyre just complains that it cannot
figure out where the attributes are getting initialized. Putting Optional here
is not a choice for us, because we want to differentiate between Optional and
strictly required attributes in our Config classes.
Reference: https://fburl.com/4cdf3akr
"""
from __future__ import annotations

from dataclasses import dataclass

from flsim.utils.async_trainer.async_example_weights import (
    EqualExampleWeightConfig,
    AsyncExampleWeightConfig,
)
from flsim.utils.async_trainer.async_staleness_weights import (
    AsyncStalenessWeightConfig,
    ConstantStalenessWeightConfig,
)
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.fl.stats import RandomVariableStatsTracker
from hydra.utils import instantiate
from omegaconf import OmegaConf


class AsyncWeight:
    def __init__(self, **kwargs) -> None:
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=AsyncWeightConfig,
            **kwargs,
        )
        # pyre-fixme[16]: `AsyncWeight` has no attribute `cfg`.
        self.example_weight = instantiate(self.cfg.example_weight)
        self.staleness_weight = instantiate(self.cfg.staleness_weight)
        self.stats = RandomVariableStatsTracker()

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        if OmegaConf.is_missing(cfg.staleness_weight, "_target_"):
            cfg.staleness_weight = ConstantStalenessWeightConfig()
        if OmegaConf.is_missing(cfg.example_weight, "_target_"):
            cfg.example_weight = EqualExampleWeightConfig()

    def weight(self, num_examples: float, staleness: int) -> float:
        weight = self.example_weight.weight(
            num_examples
        ) * self.staleness_weight.weight(staleness)
        self.stats.update(weight)
        return weight


@dataclass
class AsyncWeightConfig:
    _target_: str = fullclassname(AsyncWeight)
    _recursive_: bool = False
    staleness_weight: AsyncStalenessWeightConfig = AsyncStalenessWeightConfig()
    example_weight: AsyncExampleWeightConfig = AsyncExampleWeightConfig()



# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import math
from enum import Enum
from typing import List, Tuple


class DiversityMetricType(Enum):
    gradient_diversity = 1
    orthogonality = 2
    delta_norm_sq = 3
    sum_client_delta_norm_sq = 4
    sum_client_delta_mutual_angles = 5


class DiversityMetrics:
    """
    This class stores diversity metrics for a set of gradients -- see arXiv:1706.05699
    For a set of vectors grad f_i,

          sum || grad f_i ||_2^2          sum_of_norms
    GD =  ______________________    = ___________________
          || sum grad f_i ||_2^2          norm_of_sum

                    sum || grad f_i ||_2^2                  sum_of_norms
    ORTH =  ___________________________________ =  ____________________________
            sum{i != j} dot(grad f_i, grad f_j)     norm_of_sum - sum_of_norms


    delta_norm_sq =  norm_of_sum

    sum_client_delta_norm_sq = sum_of_norms

    sum_client_delta_mutual_angles = norm_of_sum - sum_of_norms


    For any set of n vectors, GD >= 1/n. ORTH and GD can be infinite,
    so we store the reciprocal of these quantities. This gives us good bounds:
    0 <= 1/GD <= n and -1 <= 1/ORTH <= n-1

    delta_norm_sq is the magnitude of the aggregate gradient step.
    This is a nonnegative quantity.

    sum_client_delta_norm_sq is the sum of client gradient updates.

    sum_client_delta_mutual_angles depends only on the sum of mutual angles
    between each pair client gradients. It can be positive or negative.

    The comparators will output a comparison of GD (or ORTH), not 1/GD.
    So a > b means GD(a) > GD(b), i.e. internal comparison of 1/GD(a) < 1/GD(b)

    Inputs:
    norm_of_sum >= 0
    sum_of_norms > 0
    """

    def __init__(
        self,
        norm_of_sum: float,
        sum_of_norms: float,
        diversity_metric_type: DiversityMetricType = DiversityMetricType.gradient_diversity,
    ):

        if not isinstance(diversity_metric_type, DiversityMetricType):
            raise ValueError("diversity_metric_type must be of DiversityMetricType")

        self._recpr_gradient_diversity = norm_of_sum / sum_of_norms
        self._recpr_orthogonality = (norm_of_sum - sum_of_norms) / sum_of_norms
        self._norm_of_sum = norm_of_sum
        self._sum_of_norms = sum_of_norms
        self._diversity_metric_type = diversity_metric_type

    # Provide getters but not setters, since the GD and ORTH are not actually stored.
    @property
    def gradient_diversity(self) -> float:
        if self._recpr_gradient_diversity == 0:
            return math.nan
        return 1.0 / self._recpr_gradient_diversity

    @property
    def orthogonality(self) -> float:
        if self._recpr_orthogonality == 0:
            return math.nan
        return 1.0 / self._recpr_orthogonality

    @property
    def delta_norm_sq(self) -> float:
        return self._norm_of_sum

    @property
    def sum_client_delta_norm_sq(self) -> float:
        return self._sum_of_norms

    @property
    def sum_client_delta_mutual_angles(self) -> float:
        return self._norm_of_sum - self._sum_of_norms

    @property
    def metric_value(self) -> float:
        if self._diversity_metric_type == DiversityMetricType.orthogonality:
            return self.orthogonality
        elif self._diversity_metric_type == DiversityMetricType.delta_norm_sq:
            return self.delta_norm_sq
        elif (
            self._diversity_metric_type == DiversityMetricType.sum_client_delta_norm_sq
        ):
            return self.sum_client_delta_norm_sq
        elif (
            self._diversity_metric_type
            == DiversityMetricType.sum_client_delta_mutual_angles
        ):
            return self.sum_client_delta_mutual_angles
        return self.gradient_diversity

    @property
    def diversity_metric_type(self) -> DiversityMetricType:
        return self._diversity_metric_type

    @diversity_metric_type.setter
    def diversity_metric_type(self, diversity_metric_type: DiversityMetricType):
        if not isinstance(self.diversity_metric_type, DiversityMetricType):
            raise ValueError("diversity_metric_type must be of DiversityMetricType")
        self._diversity_metric_type = diversity_metric_type

    # Metrics must not be 'None' to be equal. If LHS has a value and RHS is
    # 'None', LHS is greater. Since quantities considered are reciprocals, the
    # inequalities below are reversed from what one would expect.
    def __eq__(self, other):
        v1, v2 = self._get_metrics_of_interest(other)
        return math.isclose(v1, v2, rel_tol=1e-06)

    def __ne__(self, other):
        v1, v2 = self._get_metrics_of_interest(other)
        return v1 != v2

    def __gt__(self, other):
        v1, v2 = self._get_metrics_of_interest(other)
        return v1 < v2

    def __lt__(self, other):
        v1, v2 = self._get_metrics_of_interest(other)
        return v1 > v2

    def __ge__(self, other):
        v1, v2 = self._get_metrics_of_interest(other)
        return v1 <= v2

    def __le__(self, other):
        v1, v2 = self._get_metrics_of_interest(other)
        return v1 >= v2

    def __repr__(self):
        return "%5.5f" % self.metric_value

    def _get_metrics_of_interest(self, other) -> Tuple[float, float]:
        # For comparison, the two objects must have the same metric of interest
        assert self.diversity_metric_type == other.diversity_metric_type

        v1 = self._recpr_gradient_diversity
        v2 = other._recpr_gradient_diversity
        if self.diversity_metric_type == DiversityMetricType.orthogonality:
            v1 = self._recpr_orthogonality
            v2 = other._recpr_orthogonality
        elif self.diversity_metric_type == DiversityMetricType.delta_norm_sq:
            # Use negatives for comparison, since internal comparators are flipped
            v1 = -self.delta_norm_sq
            v2 = -other.delta_norm_sq
        elif self.diversity_metric_type == DiversityMetricType.sum_client_delta_norm_sq:
            # Use negatives for comparison, since internal comparators are flipped
            v1 = -self.sum_client_delta_norm_sq
            v2 = -other.sum_client_delta_norm_sq
        elif (
            self.diversity_metric_type
            == DiversityMetricType.sum_client_delta_mutual_angles
        ):
            # Use negatives for comparison, since internal comparators are flipped
            v1 = -self.sum_client_delta_mutual_angles
            v2 = -other.sum_client_delta_mutual_angles

        return v1, v2


class DiversityStatistics:
    def __init__(self, diversity_metrics_list: List[DiversityMetrics]):

        if len(diversity_metrics_list) == 0:
            raise ValueError("diversity_metrics_list must not be empty")

        metrics = [
            cohort_metric.metric_value for cohort_metric in diversity_metrics_list
        ]

        self.maximum_metric = max(metrics)
        self.average_metric = sum(metrics) / len(metrics)
        self.minimum_metric = min(metrics)

    def __eq__(self, other):
        close_max = math.isclose(
            self.maximum_metric, other.maximum_metric, rel_tol=1e-04
        )
        close_avg = math.isclose(
            self.average_metric, other.average_metric, rel_tol=1e-04
        )
        close_min = math.isclose(
            self.minimum_metric, other.minimum_metric, rel_tol=1e-04
        )
        return close_max and close_avg and close_min

    def __repr__(self):
        return "%5.5f,%5.5f,%5.5f" % (
            self.maximum_metric,
            self.average_metric,
            self.minimum_metric,
        )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import abc
from typing import Any, Iterable


class IFLDataLoader(abc.ABC):
    @abc.abstractmethod
    def fl_train_set(self, **kwargs) -> Iterable[Any]:
        pass

    @abc.abstractmethod
    def fl_eval_set(self, **kwargs) -> Iterable[Any]:
        pass

    @abc.abstractmethod
    def fl_test_set(self, **kwargs) -> Iterable[Any]:
        pass



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import numpy as np
import torch
from flsim.common.pytest_helper import (
    assertTrue,
    assertIsInstance,
    assertAlmostEqual,
    assertEqual,
)
from flsim.common.timeout_simulator import (
    GaussianTimeOutSimulator,
    GaussianTimeOutSimulatorConfig,
)
from flsim.utils.timing.training_duration_distribution import (
    PerExampleGaussianDurationDistributionConfig,
)
from omegaconf import OmegaConf


class TestTrainingTimeOutSimulator:
    def test_online_stat_computation_correct(self):
        timeout_simulator = GaussianTimeOutSimulator(
            **OmegaConf.structured(
                GaussianTimeOutSimulatorConfig(
                    timeout_wall_per_round=3.0,
                    fl_stopping_time=99999.0,
                    duration_distribution_generator=PerExampleGaussianDurationDistributionConfig(
                        training_duration_mean=1.0, training_duration_sd=1.0
                    ),
                ),
            )
        )
        assertTrue(isinstance(timeout_simulator, GaussianTimeOutSimulator))
        num_users = 1000
        max_sample_per_user = 20
        num_samples_per_user = [
            (user % max_sample_per_user) + 1 for user in range(num_users)
        ]
        training_time_all_users = []
        for i in range(num_users):
            sim_device_perf = timeout_simulator.simulate_per_example_training_time()
            sim_train_time = timeout_simulator.simulate_training_time(
                sim_device_perf, num_samples_per_user[i]
            )
            timeout_simulator.track_training_time_distribution(sim_train_time)
            training_time_all_users.append(sim_train_time)

        # using np.allclose to compare floats
        assertTrue(
            np.allclose(
                [timeout_simulator.sample_mean_per_user],
                [np.mean(training_time_all_users)],
            )
        )
        assertTrue(
            np.allclose(
                [timeout_simulator.sample_var_per_user],
                [np.var(training_time_all_users, ddof=1)],
            )
        )

    def test_fl_stops_small_stopping_time(self):
        """
        create a dummy "training loop" (loop through users and rounds
        without actual training) and stops the training loop by explicitly
        comparing against fl_stopping_time, versus stopping the training
        loop via internally tracked variablesin timeout_simulator and
        timeout_simulator.stop_fl() method.
        """

        timeout_simulator = GaussianTimeOutSimulator(
            **OmegaConf.structured(
                GaussianTimeOutSimulatorConfig(
                    timeout_wall_per_round=7.0,
                    fl_stopping_time=10.0,
                    duration_distribution_generator=PerExampleGaussianDurationDistributionConfig(
                        training_duration_mean=5.0, training_duration_sd=1.0
                    ),
                ),
            )
        )
        assertIsInstance(timeout_simulator, GaussianTimeOutSimulator)
        num_users = 1000
        max_sample_per_user = 20
        num_rounds = 10
        num_users_per_round = int(num_users / num_rounds)
        num_samples_per_user = [
            (user % max_sample_per_user) + 1 for user in range(num_users)
        ]
        elapsed_time_each_round = []
        torch.manual_seed(1)
        # ground-truth: maually calling private functions in timeout_simulator and
        # record all elapsed time into list for book-keeping
        for r in range(num_rounds):
            training_time_users_in_round = []
            for i in range(num_users_per_round):
                sim_device_perf = timeout_simulator.simulate_per_example_training_time()
                sim_train_time = timeout_simulator.simulate_training_time(
                    sim_device_perf, num_samples_per_user[r * num_users_per_round + i]
                )
                training_time_users_in_round.append(sim_train_time)
            elapsed_time_each_round.append(max(training_time_users_in_round))
            if sum(elapsed_time_each_round) >= timeout_simulator._fl_stopping_time:
                fl_stopping_round_ground_truth = r
                break

        torch.manual_seed(1)
        # using timeout_simulator.track_fl_elapsed_time and timeout_simulator.stop_fl()
        # to determine stopping condition
        for r in range(num_rounds):
            training_time_users_in_round = []
            for i in range(num_users_per_round):
                sim_device_perf = timeout_simulator.simulate_per_example_training_time()
                sim_train_time = timeout_simulator.simulate_training_time(
                    sim_device_perf, num_samples_per_user[r * num_users_per_round + i]
                )
                training_time_users_in_round.append(sim_train_time)

            timeout_simulator.track_fl_elapsed_time(training_time_users_in_round)
            if timeout_simulator.stop_fl():
                fl_stopping_round_simulator = r
                break

        assertAlmostEqual(
            sum(elapsed_time_each_round),
            timeout_simulator._fl_total_elapse_time,
            delta=1e-6,
        )
        assertEqual(fl_stopping_round_ground_truth, fl_stopping_round_simulator)

    def test_fl_stops_small_stopping_time_2(self):
        r"""
        training time should not be negative
        """
        timeout_simulator = GaussianTimeOutSimulator(
            **OmegaConf.structured(
                GaussianTimeOutSimulatorConfig(
                    timeout_wall_per_round=9999,
                    fl_stopping_time=99999,
                    duration_distribution_generator=PerExampleGaussianDurationDistributionConfig(
                        training_duration_mean=1.0,
                        training_duration_sd=10.0,
                        training_duration_min=0.00001,
                    ),
                ),
            )
        )
        for _ in range(1000):
            assert timeout_simulator.simulate_per_example_training_time() > 0.0



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import random
from collections import Counter
from typing import Tuple

import torch


class EM:
    """
    EM algorithm used to quantize the columns of W to minimize

                         ||W - W_quantized||^2

    Args:
        - W: weight matrix of size (n_features x n_samples)
        - n_iter: number of k-means iterations
        - n_centroids: number of centroids (size of codebook)
        - n_iter: number of E/M steps to perform
        - eps: for cluster reassignment when an empty cluster is found
        - max_tentatives for cluster reassignment when an empty cluster is found
        - verbose: print quantization error after each iteration

    Remark:
        - If one cluster is empty, the most populated cluster is split into
          two clusters.
    """

    def __init__(
        self,
        W: torch.Tensor,
        n_centroids: int = 256,
        n_iter: int = 20,
        eps: float = 1e-6,
        max_tentatives: int = 30,
        verbose: bool = False,
    ):
        self.W = W
        self.n_centroids = n_centroids
        self.n_iter = n_iter
        self.eps = eps
        self.max_tentatives = max_tentatives
        self.verbose = verbose
        self.objectives = []
        self.assignments = torch.Tensor()
        self.centroids = torch.Tensor()
        self._initialize_centroids()

    def learn(self) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Performs n_iter E/M steps.
        """
        self._initialize_centroids()
        for i in range(self.n_iter):
            self._step(i)
        return self.centroids, self.assignments

    def _initialize_centroids(self):
        """
        Initializes the centroids by sampling random columns from W.
        """

        n_features, n_samples = self.W.size()

        # centroids
        indices = torch.randint(low=0, high=n_samples, size=(self.n_centroids,)).long()
        self.centroids = self.W[:, indices].t()  # (n_centroids, n_features)

    def _step(self, i: int):
        """
        There are two standard steps for each iteration: expectation (E) and
        minimization (M). The E-step (assignment) is performed with an exhaustive
        search and the M-step (centroid computation) is performed with
        the exact solution.

        Args:
            - i: step number

        Notes:
            - The E-step heavily uses PyTorch broadcasting to speed up computations
              and reduce the memory overhead
        """

        # assignments (E-step)
        distances = self._compute_distances()  # (n_centroids, n_samples)
        self.assignments = torch.argmin(distances, dim=0)  # (n_samples)
        n_empty_resolved_clusters = self._resolve_empty_clusters()

        # centroids (M-step)
        for k in range(self.n_centroids):
            W_q = self.W[:, self.assignments == k]  # (n_features, size_of_cluster_k)
            self.centroids[k] = W_q.mean(dim=1)  # (n_features)

        # book-keeping
        obj = (self.centroids[self.assignments].t() - self.W).norm(p=2).item()
        self.objectives.append(obj)
        if self.verbose:
            print(
                f"Iteration: {i},\t"
                f"objective: {obj:.6f},\t"
                f"resolved empty clusters: {n_empty_resolved_clusters}"
            )

    def _compute_distances(self) -> torch.Tensor:
        """
        For every centroid c, computes

                          ||W - c[None, :]||_2

        Notes:
            - We rely on PyTorch's broadcasting to speed up computations
              and reduce the memory overhead
            - We use the following trick: ||a - b|| = ||a||^2 + ||b||^2 - 2 * <a, b>
        """

        W_sqr = (self.W ** 2).sum(0)  # (n_samples,)
        centroids_sqr = (self.centroids ** 2).sum(1)  # (n_centroids,)
        corr = self.centroids.mm(self.W)  # (n_centroids, n_samples)

        # return squared distances of size (n_centroids, n_samples)
        return W_sqr[None, :] + centroids_sqr[:, None] - 2 * corr

    def _resolve_empty_clusters(self) -> int:
        """
        If one cluster is empty, the most populated cluster is split into
        two clusters by shifting the respective centroids. This is done
        iteratively for a fixed number of tentatives.
        """

        # empty clusters
        counts = Counter(map(lambda x: x.item(), self.assignments))
        empty_clusters = set(range(self.n_centroids)) - set(counts.keys())
        n_empty_clusters = len(empty_clusters)

        tentatives = 0
        while len(empty_clusters) > 0:
            # given an empty cluster, find most populated cluster and split it into two
            empty_cluster = random.choice(list(empty_clusters))
            biggest_cluster = counts.most_common(1)[0][0]
            shift = torch.randn_like(self.centroids[biggest_cluster]) * self.eps
            self.centroids[empty_cluster] = self.centroids[biggest_cluster].clone()
            self.centroids[empty_cluster] += shift
            self.centroids[biggest_cluster] -= shift

            # recompute assignments
            distances = self._compute_distances()  # (n_centroids, n_samples)
            self.assignments = torch.argmin(distances, dim=0)  # (n_samples,)

            # check for empty clusters
            counts = Counter(map(lambda x: x.item(), self.assignments))
            empty_clusters = set(range(self.n_centroids)) - set(counts.keys())

            # increment tentatives
            if tentatives == self.max_tentatives:
                print(f"Could not resolve empty clusters, {len(empty_clusters)} left")
                raise EmptyClusterResolveError
            tentatives += 1

        return n_empty_clusters


class EmptyClusterResolveError(Exception):
    pass



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import copy
import logging
from dataclasses import dataclass
from enum import Enum
from typing import Optional

import torch
from flsim.channels.base_channel import IdentityChannel
from flsim.common.logger import Logger
from flsim.interfaces.model import IFLModel
from flsim.privacy.common import PrivacyBudget
from flsim.reducers.base_round_reducer import (
    ReductionType,
    RoundReducerConfig,
)
from flsim.reducers.dp_round_reducer import DPRoundReducerConfig
from flsim.reducers.weighted_dp_round_reducer import WeightedDPRoundReducerConfig
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.fl.common import FLModelParamUtils
from flsim.utils.fl.stats import ModelSequenceNumberTracker
from hydra.utils import instantiate
from omegaconf import MISSING
from torch.nn import Module as Model  # @manual


class AsyncAggregationType(Enum):
    r"""
    There are two ways in which we can aggregate local model updates
    into the global model in AsyncFL.
    Assume we're trying to aggregate updates from userA
    local_model_before_training = model sent to userA by server
    final_local_model = model after local training by userA
    global_model = model on server when userA finishes training

    new_global_model = global_model - delta*lr
    The difference is in how `delta` is computed
    a) fed_buff_aggregation: delta = local_model_before_training - final_local_model
    b) fed_async_aggregation: delta = global_model - final_local_model

    Literature uses fed_async_aggregation (https://arxiv.org/abs/1903.03934)
    We find empirically that fed_buff_aggregation performs better
    """

    fed_buff_aggregation = 1
    fed_async_aggregation = 2

    @staticmethod
    def from_str(name: str):
        name_lower = name.lower()
        names = [e.name for e in AsyncAggregationType]
        assert name_lower in names, "Unknown async aggregation type:" + name
        return AsyncAggregationType[name_lower]


class AsyncAggregator:
    r"""
    Implements Asynchronous Federated Learning
        Input: local_model_before_training, final_local_model, global_model, wt
        Output: global_model = global_model - lr*delta, where
            delta = (local_model_before_training - final_local_model)
            lr: original_lr*wt (higher wt = higher learning rate)
            original_lr: lr given in config.
                smaller original_lr: small changes in global_model
                higher original_lr: large changes in global_model
                Eg: lr=1 means keep global_model=final_local_model (when wt=1)
        TODO: add adaptive learning rate based on staleness of gradient
    Attributes:

    Internal attributes:
        optimizer (torch.optim.Optimizer): an optimizer kept
        internally to prevent optimizer creation on every epoch
    """

    logger: logging.Logger = Logger.get_logger("AsyncAggregator")

    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: Optional[IdentityChannel] = None,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=AsyncAggregatorConfig,
            **kwargs,
        )

        self.optimizer = create_optimizer_for_async_aggregator(
            # pyre-fixme[16]: `AsyncAggregator` has no attribute `cfg`.
            self.cfg,
            global_model.fl_get_module(),
        )

        if self.cfg.aggregation_type not in [
            AsyncAggregationType.fed_buff_aggregation,
            AsyncAggregationType.fed_async_aggregation,
        ]:
            raise AssertionError(
                f"Unknown AsyncAggregationType:{self.cfg.aggregation_type}"
            )

        self.orig_lr = self.optimizer.param_groups[0]["lr"]
        self._global_model: IFLModel = global_model
        self._reconstructed_grad: IFLModel = copy.deepcopy(self._global_model)
        # there is no concept of a round in async, hence round reducer is not tied to a round
        self.reducer = instantiate(
            self.cfg.reducer,
            global_model=self._global_model,
            channel=channel,
            # TODO: ngjhn passing this here for DP reducer instantiation
            # will fix in T80906721
            num_users_per_round=self.cfg.num_users_per_round,
            total_number_of_users=self.cfg.total_number_of_users,
        )
        self.seqnum_tracker = ModelSequenceNumberTracker()

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    @property
    def global_model(self) -> IFLModel:
        return self._global_model

    def set_num_total_users(self, num_total_users: int) -> None:
        # TODO: ngjhn this is ugly and we will fix this by moving
        # num_total_users to the constructor of TrainerBase T80906721
        self.reducer.set_num_total_users(num_total_users)

    def model_staleness(self, model_seqnum: int):
        seqnum_diff = self.seqnum_tracker.get_staleness_and_update_stats(model_seqnum)
        return seqnum_diff

    def zero_grad(self):
        r"""
        Reset the reducer and optimizer to prepare to update
        the global model
        """
        self.reducer.reset(ref_model=self._global_model)
        self.optimizer.zero_grad()

    @property
    def privacy_budget(self) -> Optional[PrivacyBudget]:
        return self.reducer.privacy_budget if self.is_private else None

    @property
    def global_seqnum(self):
        return self.seqnum_tracker.current_seqnum

    @property
    def is_private(self):
        return self.cfg.reducer._target_ in {
            DPRoundReducerConfig._target_,
            WeightedDPRoundReducerConfig._target_,
        }

    def on_client_training_end(
        self,
        client_delta: IFLModel,
        final_local_model: IFLModel,
        weight: float,
    ) -> bool:
        """
        Collects the client update and updates the global model

        Note:
            Async updates the global model on every client hence return value is always true

        Args:
            client_delta (IFLModel): the difference between the client's before and after training
            final_local_model (IFLModel): client's model after local training
            weight (float): client's update weight

        Returns:
            bool: whether or not the global model was updated
        """
        client_delta = self.reducer.receive_through_channel(client_delta)
        # pyre-fixme[16]: `AsyncAggregator` has no attribute `cfg`.
        if self.cfg.aggregation_type == AsyncAggregationType.fed_buff_aggregation:
            FLModelParamUtils.set_gradient(
                model=self._global_model.fl_get_module(),
                reference_gradient=client_delta.fl_get_module(),
            )
        else:
            FLModelParamUtils.subtract_model(
                minuend=self._global_model.fl_get_module(),
                subtrahend=final_local_model.fl_get_module(),
                difference=self._reconstructed_grad.fl_get_module(),
            )
            FLModelParamUtils.set_gradient(
                model=self._global_model.fl_get_module(),
                reference_gradient=self._reconstructed_grad.fl_get_module(),
            )

        self._step_with_modified_lr(lr_normalizer=weight)
        return True

    def on_training_epoch_end(self) -> bool:
        """
        Base method for end of training epoch
        Returns whether some client updates were aggregated into global model
          (In k-async, up to (k-1) client updates may not have been aggregated
          at the end of the epoch, and need to be handled explicitly)
        """
        pass

    def _step_with_modified_lr(self, lr_normalizer: float):
        r"""
        Updates the learning rate based on the weight and
        increments global seqnum

        original_lr: lr given in config
        lr: original_lr * weight (higher weight = higher learning rate)

        smaller original_lr: small changes in global_model
        higher original_lr: large changes in global_model
        Eg: lr=1 means keep global_model=final_local_model (when wt=1)
        """
        # TODO: for optimizers that don't use momentum or adaptive learning rate,
        # scaling LR directly is optimal
        # For optimizers like Adam, or SGD+Momentum, add optimal weight-based scaling
        # (open research problem)
        for param_group in self.optimizer.param_groups:
            param_group["lr"] = self.orig_lr * lr_normalizer
        self.optimizer.step()
        self.seqnum_tracker.increment()


class FedAdamAsyncAggregator(AsyncAggregator):
    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: Optional[IdentityChannel] = None,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FedAdamAsyncAggregatorConfig,
            **kwargs,
        )

        super().__init__(global_model=global_model, channel=channel, **kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass


class FedAvgWithLRAsyncAggregator(AsyncAggregator):
    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: Optional[IdentityChannel] = None,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FedAvgWithLRAsyncAggregatorConfig,
            **kwargs,
        )

        super().__init__(global_model=global_model, channel=channel, **kwargs)


class FedAvgWithLRWithMomentumAsyncAggregator(FedAvgWithLRAsyncAggregator):
    """Implements the correct version of momentum"""

    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: Optional[IdentityChannel] = None,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FedAvgWithLRWithMomentumAsyncAggregatorConfig,
            **kwargs,
        )

        super().__init__(global_model=global_model, channel=channel, **kwargs)
        self.orig_momentum = self.optimizer.param_groups[0]["momentum"]
        assert (
            # pyre-fixme[16]: `FedAvgWithLRWithMomentumAsyncAggregator` has no attribute
            #  `cfg`.
            self.cfg.aggregation_type
            == AsyncAggregationType.fed_buff_aggregation
        ), "Only delta direction is supported by "
        f"{FedAvgWithLRWithMomentumAsyncAggregatorConfig.__class__.__name__}"

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def on_client_training_end(
        self,
        client_delta: IFLModel,
        final_local_model: IFLModel,
        weight: float,
    ) -> bool:
        # TODO: better if this assert fires at config creation time
        assert (
            weight >= 0 and weight <= 1
        ), f"{FedAvgWithLRWithMomentumAsyncAggregatorConfig.__class__.__name__}"
        "only supports weights between 0 and 1"

        # scale delta by weight
        FLModelParamUtils.multiply_model_by_weight(
            model=client_delta.fl_get_module(),
            weight=weight,
            model_to_save=self._reconstructed_grad.fl_get_module(),
        )
        FLModelParamUtils.set_gradient(
            model=self._global_model.fl_get_module(),
            reference_gradient=self._reconstructed_grad.fl_get_module(),
        )

        # change momentum parameter
        momentum = self.orig_momentum + ((1 - self.orig_momentum) * (1 - weight))
        for param_group in self.optimizer.param_groups:
            param_group["momentum"] = momentum
        self._step_with_modified_lr(lr_normalizer=weight)
        return True


class HybridAggregator(AsyncAggregator):
    r"""
    Aggregator for Hybrid-FL where client update and global update are decoupled

    Keeps track of number clients reported and take a global step after reaching the
    threshold set by the config
    """
    logger: logging.Logger = Logger.get_logger("HybridAggregator")

    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: Optional[IdentityChannel] = None,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=HybridAggregatorConfig,
            **kwargs,
        )

        super().__init__(global_model=global_model, channel=channel, **kwargs)
        # pyre-fixme[16]: `HybridAggregator` has no attribute `cfg`.
        if self.cfg.aggregation_type != AsyncAggregationType.fed_buff_aggregation:
            raise ValueError("Hybrid Aggregator only supports delta direction")
        self.num_clients_reported = 0

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def zero_grad(self):
        r"""
        Zero'd out the grads if it's the first update or if
        reaching buffer_size otherwise no-op
        """
        if self.num_clients_reported == 0 or self.should_update_global_model():
            super().zero_grad()

    def on_client_training_end(
        self,
        client_delta: IFLModel,
        final_local_model: IFLModel,
        weight: float,
    ) -> bool:
        """
        Collects client update and update global model if reaching buffer_size

        Args:
            client_delta (IFLModel): the difference between the client's before and after training
            final_local_model (IFLModel): client's model after local training
            weight (float): client's update weight

        Returns:
            bool: whether or not the global model was updated
        """
        # total_delta += delta
        self._collect_client_update(update=client_delta, weight=weight)
        if self.should_update_global_model():
            self._update_global_model()
            return True
        return False

    def should_update_global_model(self) -> bool:
        # pyre-fixme[16]: `HybridAggregator` has no attribute `cfg`.
        return self.num_clients_reported >= self.cfg.buffer_size

    def on_training_epoch_end(self) -> bool:
        """
        Updates the global model in case when
        there are remaining clients who didn't get aggregated
        into the global model at the end of an epoch.
        Return value:
          True if there were any such clients with pending updates. In this case,
            the global model was updated.
          False if there were no such clients. Global model was not update
        """
        if self.num_clients_reported != 0:
            self._update_global_model()
            return True
        return False

    def _update_global_model(self):
        total_delta, _ = self.reducer.reduce()
        FLModelParamUtils.set_gradient(
            model=self._global_model.fl_get_module(), reference_gradient=total_delta
        )
        self._step_with_modified_lr(lr_normalizer=1.0)
        self.num_clients_reported = 0

    def _collect_client_update(self, update: IFLModel, weight: float) -> None:
        """
        Collects update from one client and aggregtes it internally.
        reduced model = reduced model + update * weight
        """
        self.reducer.collect_update(delta=update, weight=weight)
        self.num_clients_reported += 1


class FedAvgWithLRHybridAggregator(HybridAggregator):
    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: Optional[IdentityChannel] = None,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FedAvgWithLRHybridAggregatorConfig,
            **kwargs,
        )

        super().__init__(global_model=global_model, channel=channel, **kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass


class FedAdamHybridAggregator(HybridAggregator):
    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: Optional[IdentityChannel] = None,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FedAdamHybridAggregatorConfig,
            **kwargs,
        )

        super().__init__(global_model=global_model, channel=channel, **kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass


def create_optimizer_for_async_aggregator(config: AsyncAggregatorConfig, model: Model):
    if config._target_ in {
        FedAvgWithLRAsyncAggregatorConfig._target_,
        FedAvgWithLRHybridAggregatorConfig._target_,
        FedAvgWithLRWithMomentumAsyncAggregatorConfig._target_,
    }:
        return torch.optim.SGD(
            model.parameters(),
            # pyre-fixme[16]: `AsyncAggregatorConfig` has no attribute `lr`.
            lr=config.lr,
            # pyre-fixme[16]: `AsyncAggregatorConfig` has no attribute `momentum`.
            momentum=config.momentum,
        )
    elif config._target_ in {
        FedAdamAsyncAggregatorConfig._target_,
        FedAdamHybridAggregatorConfig._target_,
    }:
        return torch.optim.Adam(
            model.parameters(),
            lr=config.lr,
            # pyre-fixme[16]: `AsyncAggregatorConfig` has no attribute `weight_decay`.
            weight_decay=config.weight_decay,
            # pyre-fixme[16]: `AsyncAggregatorConfig` has no attribute `beta1`.
            # pyre-fixme[16]: `AsyncAggregatorConfig` has no attribute `beta2`.
            betas=(config.beta1, config.beta2),
            # pyre-fixme[16]: `AsyncAggregatorConfig` has no attribute `eps`.
            eps=config.eps,
        )


@dataclass
class AsyncAggregatorConfig:
    _target_: str = MISSING
    _recursive_: bool = False
    aggregation_type: AsyncAggregationType = AsyncAggregationType.fed_buff_aggregation
    # reducer to collect client updates, there is no concept of a round
    # in async, hence round reducer is not tied to a round
    reducer: RoundReducerConfig = RoundReducerConfig(
        reduction_type=ReductionType.WEIGHTED_SUM
    )
    # TODO: ngjhn passing this here for DP reducer instantiation
    # will fix in T80906721
    num_users_per_round: int = 1
    total_number_of_users: int = 10000000000


@dataclass
class FedAvgWithLRAsyncAggregatorConfig(AsyncAggregatorConfig):
    _target_: str = fullclassname(FedAvgWithLRAsyncAggregator)
    lr: float = 0.001
    momentum: float = 0.0


@dataclass
class FedAvgWithLRWithMomentumAsyncAggregatorConfig(FedAvgWithLRAsyncAggregatorConfig):
    _target_: str = fullclassname(FedAvgWithLRWithMomentumAsyncAggregator)


@dataclass
class FedAdamAsyncAggregatorConfig(AsyncAggregatorConfig):
    _target_: str = fullclassname(FedAdamAsyncAggregator)
    lr: float = 0.001
    weight_decay: float = 0.00001
    beta1: float = 0.9
    beta2: float = 0.999
    eps: float = 1e-8


@dataclass
class HybridAggregatorConfig(AsyncAggregatorConfig):
    # number of clients to collect before taking a global step
    buffer_size: int = 1


@dataclass
class FedAvgWithLRHybridAggregatorConfig(HybridAggregatorConfig):
    _target_: str = fullclassname(FedAvgWithLRHybridAggregator)
    lr: float = 0.001
    momentum: float = 0.0


@dataclass
class FedAdamHybridAggregatorConfig(HybridAggregatorConfig):
    _target_: str = fullclassname(FedAdamHybridAggregator)
    lr: float = 0.001
    weight_decay: float = 0.00001
    beta1: float = 0.9
    beta2: float = 0.999
    eps: float = 1e-8



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

from collections import OrderedDict
from dataclasses import dataclass

from flsim.channels.base_channel import (
    IdentityChannel,
    FLChannelConfig,
    Message,
)
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg


class HalfPrecisionChannel(IdentityChannel):
    """
    Implements a channel that emulates upload of model from client
    to server using half precision (fp16). We emulate this by
    casting successively to half() and back to float(). This way,
    the rest of the training is transparent for aggreagators, reducers,
    trainers and so on.

    Note that model.half().float() is *not* in general a no-operation,
    since we lose a bit of information when casting to half (and that
    is the desired behavior). That's why the corresponding channel_test
    is done with a larger `atol` in the direction client->server.

    We do not override the method ``_during_transmission_client_to_server``
    from IdentityChannel since it can be applied here without modification to
    measure the message size. Inded, it relies on ``element_size()``, which is,
    as expected, 2 bytes for half precision and 4 bytes for fp32.
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=HalfPrecisionChannelConfig,
            **kwargs,
        )
        super().__init__(**kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _on_client_before_transmission(self, message: Message) -> Message:
        """
        Here we cast all parameters to half precision. We copy the
        state dict since the tensor format changes.
        """
        message.populate_state_dict()
        new_state_dict = OrderedDict()
        for name, param in message.model_state_dict.items():
            new_state_dict[name] = param.data.half()

        message.model_state_dict = new_state_dict
        return message

    def _on_server_before_transmission(self, message: Message) -> Message:
        """Populate state dict with model before transmission"""
        message.populate_state_dict()
        return message

    def _on_server_after_reception(self, message: Message) -> Message:
        """
        We decompress the message by casting back all parameters
        to full precision (fp32). Note that this is not a no-op
        since there is some loss in precision when casting to half().
        We copy the state dict since the tensor format changes.
        """
        new_state_dict = OrderedDict()
        for name, param in message.model_state_dict.items():
            new_state_dict[name] = param.data.float()

        message.model_state_dict = new_state_dict
        message.update_model_()
        return message


@dataclass
class HalfPrecisionChannelConfig(FLChannelConfig):
    _target_: str = fullclassname(HalfPrecisionChannel)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from flsim.utils.fl.stats import (
    AverageType,
)
from flsim.utils.fl.target_metric import TargetMetricTracker, TargetMetricDirection
from libfb.py import testutil


class TargetMetricTestTest(testutil.BaseFacebookTestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_target_metric_optimize_for_max(self) -> None:
        """
        Test if target tracker returns true when the sliding window returns
        true when optimizing for max value (e.g accuracy)
        For example, target = 90, window_size = 3
        eval accuracies = [80, 81, 89, 90, 95]
        [80, 81, 89] -> false
        [81, 89, 90] -> false
        the window [89, 90, 95] -> true
        """
        metrics = [80, 81, 89, 90, 95]
        target_value = 90
        for average_type in [AverageType.SMA, AverageType.EMA]:
            target_tracker = TargetMetricTracker(
                target_value=target_value,
                window_size=3,
                average_type=average_type,
                direction=TargetMetricDirection.MAX,
            )
            for metric in metrics[:-1]:
                self.assertFalse(target_tracker.update_and_check_target(metric))
                self.assertLess(target_tracker.mean, target_value)

            self.assertTrue(target_tracker.update_and_check_target(metrics[-1]))
            self.assertGreater(target_tracker.mean, target_value)

    def test_target_metric_optimize_for_min(self) -> None:
        """
        Test if target tracker returns true when the sliding window returns
        true when optimizing for min value (e.g loss)
        For example, target = 0.1, window_size = 3
        eval loss = [0.5, 0.4, 0.15, 0.04, 0.1
        [0.5, 0.4, 0.15] -> false
        [0.4, 0.15, 0.04] -> false
        the window [0.15, 0.04, 0.1] -> true
        """
        metrics = [0.5, 0.4, 0.15, 0.04, 0.1]
        target_value = 0.1
        for average_type in [AverageType.SMA, AverageType.EMA]:
            target_tracker = TargetMetricTracker(
                target_value=target_value,
                window_size=3,
                average_type=average_type,
                direction=TargetMetricDirection.MIN,
            )
            for metric in metrics[:-1]:
                self.assertFalse(target_tracker.update_and_check_target(metric))
                self.assertGreater(target_tracker.mean, target_value)

            self.assertTrue(target_tracker.update_and_check_target(metrics[-1]))
            self.assertLess(target_tracker.mean, target_value)
